import sys
from pathlib import Path


def test_fix_non_existing_problem(project, pdm):
    result = pdm(["fix", "non-existing"], obj=project)
    assert result.exit_code == 1


def test_fix_individual_problem(project, pdm):
    project._saved_python = None
    old_config = project.root / ".pdm.toml"
    old_config.write_text(f'[python]\nuse_pyenv = false\npath = "{Path(sys.executable).as_posix()}"\n')
    pdm(["fix", "project-config"], obj=project, strict=True)
    assert not old_config.exists()


def test_show_fix_command(project, pdm):
    old_config = project.root / ".pdm.toml"
    old_config.write_text(f'[python]\nuse_pyenv = false\npath = "{Path(sys.executable).as_posix()}"\n')
    result = pdm(["info"], obj=project)
    assert "Run pdm fix to fix all" in result.stderr

    result = pdm(["fix", "-h"], obj=project)
    assert "Run pdm fix to fix all" not in result.stderr


def test_show_fix_command_global_project(core, pdm, project_no_init):
    project = core.create_project(None, True, project_no_init.global_config.config_file)
    old_config = project.root / ".pdm.toml"
    old_config.write_text(f'[python]\nuse_pyenv = false\npath = "{Path(sys.executable).as_posix()}"\n')
    result = pdm(["info"], obj=project)
    assert "Run pdm fix -g to fix all" in result.stderr

    result = pdm(["fix", "-h"], obj=project)
    assert "Run pdm fix -g to fix all" not in result.stderr


def test_fix_project_config(project, pdm):
    project._saved_python = None
    old_config = project.root / ".pdm.toml"
    old_config.write_text(f'[python]\nuse_pyenv = false\npath = "{Path(sys.executable).as_posix()}"\n')
    assert project.project_config["python.use_pyenv"] is False
    assert project._saved_python == Path(sys.executable).as_posix()
    pdm(["fix"], obj=project, strict=True)
    assert not old_config.exists()
    assert project.root.joinpath("pdm.toml").read_text() == "[python]\nuse_pyenv = false\n"
    assert project.root.joinpath(".pdm-python").read_text().strip() == Path(sys.executable).as_posix()



from itertools import chain
from unittest import mock

import pytest

from pdm import signals
from pdm.models.candidates import Candidate
from pdm.models.repositories import Package
from pdm.models.requirements import Requirement


def test_post_init_signal(project_no_init, pdm):
    mock_handler = mock.Mock()
    with signals.post_init.connected_to(mock_handler):
        result = pdm(["init"], input="\n\n\n\n\n\n\n\n", obj=project_no_init)
        assert result.exit_code == 0
    mock_handler.assert_called_once_with(project_no_init, hooks=mock.ANY)


@pytest.mark.usefixtures("working_set")
def test_post_lock_and_install_signals(project, pdm):
    pre_lock = signals.pre_lock.connect(mock.Mock(), weak=False)
    post_lock = signals.post_lock.connect(mock.Mock(), weak=False)
    pre_install = signals.pre_install.connect(mock.Mock(), weak=False)
    post_install = signals.post_install.connect(mock.Mock(), weak=False)
    pdm(["add", "requests"], obj=project, strict=True)
    signals.pre_lock.disconnect(pre_lock)
    signals.post_lock.disconnect(post_lock)
    signals.pre_install.disconnect(pre_install)
    signals.post_install.disconnect(post_install)
    for mocker in (pre_lock, post_lock, pre_install, post_install):
        mocker.assert_called_once()


@pytest.mark.usefixtures("working_set")
def test_lock_and_install_signals_injection_with_add(project, pdm):
    pre_lock = signals.pre_lock.connect(mock.Mock(), weak=False)
    post_lock = signals.post_lock.connect(mock.Mock(), weak=False)
    pre_install = signals.pre_install.connect(mock.Mock(), weak=False)
    post_install = signals.post_install.connect(mock.Mock(), weak=False)
    pdm(["add", "requests"], obj=project, strict=True)
    signals.pre_lock.disconnect(pre_lock)
    signals.post_lock.disconnect(post_lock)
    signals.pre_install.disconnect(pre_install)
    signals.post_install.disconnect(post_install)

    assert isinstance(pre_lock.call_args.kwargs["requirements"], list)
    assert all(isinstance(e, Requirement) for e in pre_lock.call_args.kwargs["requirements"])
    assert len(pre_lock.call_args.kwargs["requirements"]) == 1

    assert isinstance(post_lock.call_args.kwargs["resolution"], dict)
    assert all(isinstance(e, Candidate) for e in chain.from_iterable(post_lock.call_args.kwargs["resolution"].values()))
    assert len(post_lock.call_args.kwargs["resolution"]) == 5

    assert isinstance(pre_install.call_args.kwargs["packages"], list)
    assert all(isinstance(e, Package) for e in pre_install.call_args.kwargs["packages"])
    assert len(pre_install.call_args.kwargs["packages"]) == 5

    assert isinstance(post_install.call_args.kwargs["packages"], list)
    assert all(isinstance(e, Package) for e in post_install.call_args.kwargs["packages"])
    assert len(post_install.call_args.kwargs["packages"]) == 5


@pytest.mark.usefixtures("working_set")
def test_lock_and_install_signals_injection_with_install(project, pdm):
    project.add_dependencies(["requests"])

    pre_lock = signals.pre_lock.connect(mock.Mock(), weak=False)
    post_lock = signals.post_lock.connect(mock.Mock(), weak=False)
    pre_install = signals.pre_install.connect(mock.Mock(), weak=False)
    post_install = signals.post_install.connect(mock.Mock(), weak=False)
    pdm(["install"], obj=project, strict=True)
    signals.pre_lock.disconnect(pre_lock)
    signals.post_lock.disconnect(post_lock)
    signals.pre_install.disconnect(pre_install)
    signals.post_install.disconnect(post_install)

    assert isinstance(pre_lock.call_args.kwargs["requirements"], list)
    assert all(isinstance(e, Requirement) for e in pre_lock.call_args.kwargs["requirements"])
    assert len(pre_lock.call_args.kwargs["requirements"]) == 1

    assert isinstance(post_lock.call_args.kwargs["resolution"], dict)
    assert all(isinstance(e, Candidate) for e in chain.from_iterable(post_lock.call_args.kwargs["resolution"].values()))
    assert len(post_lock.call_args.kwargs["resolution"]) == 5

    assert isinstance(pre_install.call_args.kwargs["packages"], list)
    assert all(isinstance(e, Package) for e in pre_install.call_args.kwargs["packages"])
    assert len(pre_install.call_args.kwargs["packages"]) == 5

    assert isinstance(post_install.call_args.kwargs["packages"], list)
    assert all(isinstance(e, Package) for e in post_install.call_args.kwargs["packages"])
    assert len(post_install.call_args.kwargs["packages"]) == 5


@pytest.mark.usefixtures("working_set")
def test_lock_signals_injection_with_update(project, pdm):
    project.add_dependencies(["requests"])

    pre_lock = signals.pre_lock.connect(mock.Mock(), weak=False)
    post_lock = signals.post_lock.connect(mock.Mock(), weak=False)
    pdm(["update"], obj=project, strict=True)
    signals.pre_lock.disconnect(pre_lock)
    signals.post_lock.disconnect(post_lock)

    assert isinstance(pre_lock.call_args.kwargs["requirements"], list)
    assert all(isinstance(e, Requirement) for e in pre_lock.call_args.kwargs["requirements"])
    assert len(pre_lock.call_args.kwargs["requirements"]) == 1

    assert isinstance(post_lock.call_args.kwargs["resolution"], dict)
    assert all(isinstance(e, Candidate) for e in chain.from_iterable(post_lock.call_args.kwargs["resolution"].values()))
    assert len(post_lock.call_args.kwargs["resolution"]) == 5



import json
import os
import subprocess
import textwrap
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from pdm import termui
from pdm.cli import actions
from pdm.cli.utils import get_pep582_path
from pdm.utils import cd


@pytest.fixture
def _args(project):
    (project.root / "args.py").write_text(
        textwrap.dedent(
            """
            import os
            import sys
            name = sys.argv[1]
            args = ", ".join(sys.argv[2:])
            print(f"{name} CALLED with {args}" if args else f"{name} CALLED")
            """
        )
    )


def test_pep582_launcher_for_python_interpreter(project, local_finder, pdm):
    project.root.joinpath("main.py").write_text("import first;print(first.first([0, False, 1, 2]))\n")
    result = pdm(["add", "first"], obj=project)
    assert result.exit_code == 0, result.stderr
    env = os.environ.copy()
    env.update({"PYTHONPATH": get_pep582_path(project)})
    output = subprocess.check_output(
        [str(project.python.executable), str(project.root.joinpath("main.py"))],
        env=env,
    )
    assert output.decode().strip() == "1"


def test_auto_isolate_site_packages(project, pdm):
    env = os.environ.copy()
    env.update({"PYTHONPATH": get_pep582_path(project)})
    proc = subprocess.run(
        [str(project.python.executable), "-c", "import sys;print(sys.path, sep='\\n')"],
        env=env,
        capture_output=True,
        text=True,
        cwd=str(project.root),
        check=True,
    )
    assert any("site-packages" in path for path in proc.stdout.splitlines())

    result = pdm(
        ["run", "python", "-c", "import sys;print(sys.path, sep='\\n')"],
        obj=project,
        strict=True,
    )
    assert not any("site-packages" in path for path in result.stdout.splitlines())


def test_run_with_site_packages(project, pdm):
    project.pyproject.settings["scripts"] = {
        "foo": {
            "cmd": ["python", "-c", "import sys;print(sys.path, sep='\\n')"],
            "site_packages": True,
        }
    }
    project.pyproject.write()
    result = pdm(
        [
            "run",
            "--site-packages",
            "python",
            "-c",
            "import sys;print(sys.path, sep='\\n')",
        ],
        obj=project,
    )
    assert result.exit_code == 0
    result = pdm(["run", "foo"], obj=project)
    assert result.exit_code == 0


def test_run_command_not_found(pdm):
    result = pdm(["run", "foobar"])
    assert "Command 'foobar' is not found in your PATH." in result.stderr
    assert result.exit_code == 1


def test_run_pass_exit_code(pdm):
    result = pdm(["run", "python", "-c", "1/0"])
    assert result.exit_code == 1


def test_run_cmd_script(project, pdm):
    project.pyproject.settings["scripts"] = {"test_script": "python -V"}
    project.pyproject.write()
    result = pdm(["run", "test_script"], obj=project)
    assert result.exit_code == 0


def test_run_cmd_script_with_array(project, pdm):
    project.pyproject.settings["scripts"] = {"test_script": ["python", "-c", "import sys; sys.exit(22)"]}
    project.pyproject.write()
    result = pdm(["run", "test_script"], obj=project)
    assert result.exit_code == 22


def test_run_script_pass_project_root(project, pdm, capfd):
    project.pyproject.settings["scripts"] = {
        "test_script": [
            "python",
            "-c",
            "import os;print(os.getenv('PDM_PROJECT_ROOT'))",
        ]
    }
    project.pyproject.write()
    capfd.readouterr()
    result = pdm(["run", "test_script"], obj=project)
    assert result.exit_code == 0
    out, _ = capfd.readouterr()
    assert Path(out.strip()) == project.root


def test_run_shell_script(project, pdm):
    project.pyproject.settings["scripts"] = {
        "test_script": {
            "shell": "echo hello > output.txt",
            "help": "test it won't fail",
        }
    }
    project.pyproject.write()
    with cd(project.root):
        result = pdm(["run", "test_script"], obj=project)
    assert result.exit_code == 0
    assert (project.root / "output.txt").read_text().strip() == "hello"


def test_run_script_with_relative_path(project, pdm, capfd):
    if os.name == "nt":
        (project.root / "test_script.bat").write_text("@echo Hello\n")
    else:
        (project.root / "test_script.sh").write_text("#!/bin/bash\necho Hello\n")
        (project.root / "test_script.sh").chmod(0o755)
    with cd(project.root):
        pdm(["run", "./test_script.bat" if os.name == "nt" else "./test_script.sh"], obj=project, strict=True)
    out, _ = capfd.readouterr()
    assert out.strip() == "Hello"


def test_run_non_existing_local_script(project, pdm):
    with cd(project.root):
        result = pdm(["run", "./test_script.sh"], obj=project)
    assert result.exit_code != 0
    assert "not a valid executable" in result.stderr


@pytest.mark.parametrize(
    "args,expected",
    (
        pytest.param(["hello"], "ok hello", id="with-args"),
        pytest.param([], "ok", id="without-args"),
    ),
)
def test_run_shell_script_with_args_placeholder(project, pdm, args, expected):
    project.pyproject.settings["scripts"] = {
        "test_script": {
            "shell": "echo ok {args} > output.txt",
            "help": "test it won't fail",
        }
    }
    project.pyproject.write()
    with cd(project.root):
        result = pdm(["run", "test_script", *args], obj=project)
    assert result.exit_code == 0
    assert (project.root / "output.txt").read_text().strip() == expected


@pytest.mark.parametrize(
    "args,expected",
    (
        pytest.param(["hello"], "hello", id="with-args"),
        pytest.param([], "default", id="with-default"),
    ),
)
def test_run_shell_script_with_args_placeholder_with_default(project, pdm, args, expected):
    project.pyproject.settings["scripts"] = {
        "test_script": {
            "shell": "echo {args:default} > output.txt",
            "help": "test it won't fail",
        }
    }
    project.pyproject.write()
    with cd(project.root):
        result = pdm(["run", "test_script", *args], obj=project)
    assert result.exit_code == 0
    assert (project.root / "output.txt").read_text().strip() == expected


def test_run_call_script(project, pdm):
    (project.root / "test_script.py").write_text(
        textwrap.dedent(
            """
            import argparse
            import sys

            def main(argv=None):
                parser = argparse.ArgumentParser()
                parser.add_argument("-c", "--code", type=int)
                args = parser.parse_args(argv)
                sys.exit(args.code)
            """
        )
    )
    project.pyproject.settings["scripts"] = {
        "test_script": {"call": "test_script:main"},
        "test_script_with_args": {"call": "test_script:main(['-c', '9'])"},
    }
    project.pyproject.write()
    with cd(project.root):
        result = pdm(["run", "test_script", "-c", "8"], obj=project)
        assert result.exit_code == 8

        result = pdm(["run", "test_script_with_args"], obj=project)
        assert result.exit_code == 9


def test_run_script_with_extra_args(project, pdm, capfd):
    (project.root / "test_script.py").write_text(
        textwrap.dedent(
            """
            import sys
            print(*sys.argv[1:], sep='\\n')
            """
        )
    )
    project.pyproject.settings["scripts"] = {"test_script": "python test_script.py"}
    project.pyproject.write()
    with cd(project.root):
        pdm(["run", "test_script", "-a", "-b", "-c"], obj=project)
    out, _ = capfd.readouterr()
    assert out.splitlines()[-3:] == ["-a", "-b", "-c"]


@pytest.mark.parametrize(
    "args,expected",
    (
        pytest.param(["-a", "-b", "-c"], ["-a", "-b", "-c", "-x"], id="with-args"),
        pytest.param([], ["-x"], id="without-args"),
    ),
)
@pytest.mark.parametrize(
    "script",
    (
        pytest.param("python test_script.py {args} -x", id="as-str"),
        pytest.param(["python", "test_script.py", "{args}", "-x"], id="as-list"),
    ),
)
def test_run_script_with_args_placeholder(project, pdm, capfd, script, args, expected):
    (project.root / "test_script.py").write_text(
        textwrap.dedent(
            """
            import sys
            print(*sys.argv[1:], sep='\\n')
            """
        )
    )
    project.pyproject.settings["scripts"] = {"test_script": script}
    project.pyproject.write()
    with cd(project.root):
        pdm(["run", "-v", "test_script", *args], obj=project)
    out, _ = capfd.readouterr()
    assert out.strip().splitlines()[1:] == expected


@pytest.mark.parametrize(
    "args,expected",
    (
        pytest.param(["-a", "-b", "-c"], ["-a", "-b", "-c", "-x"], id="with-args"),
        pytest.param([], ["--default", "--value", "-x"], id="default"),
    ),
)
@pytest.mark.parametrize(
    "script",
    (
        pytest.param("python test_script.py {args:--default --value} -x", id="as-str"),
        pytest.param(["python", "test_script.py", "{args:--default --value}", "-x"], id="as-list"),
    ),
)
def test_run_script_with_args_placeholder_with_default(project, pdm, capfd, script, args, expected):
    (project.root / "test_script.py").write_text(
        textwrap.dedent(
            """
            import sys
            print(*sys.argv[1:], sep='\\n')
            """
        )
    )
    project.pyproject.settings["scripts"] = {"test_script": script}
    project.pyproject.write()
    with cd(project.root):
        pdm(["run", "-v", "test_script", *args], obj=project)
    out, _ = capfd.readouterr()
    assert out.strip().splitlines()[1:] == expected


def test_run_shell_script_with_pdm_placeholder(project, pdm):
    project.pyproject.settings["scripts"] = {
        "test_script": {
            "shell": "{pdm} -V > output.txt",
            "help": "test it won't fail",
        }
    }
    project.pyproject.write()
    with cd(project.root):
        result = pdm(["run", "test_script"], obj=project)
    assert result.exit_code == 0
    assert (project.root / "output.txt").read_text().strip().startswith("PDM, version")


def test_run_expand_env_vars(project, pdm, capfd, monkeypatch):
    (project.root / "test_script.py").write_text("import os; print(os.getenv('FOO'))")
    project.pyproject.settings["scripts"] = {
        "test_cmd": 'python -c "foo, bar = 0, 1;print(${FOO})"',
        "test_cmd_no_expand": "python -c 'print(${FOO})'",
        "test_script": "python test_script.py",
        "test_cmd_array": ["python", "test_script.py"],
        "test_shell": {"shell": "echo %FOO%" if os.name == "nt" else "echo $FOO"},
    }
    project.pyproject.write()
    capfd.readouterr()
    with cd(project.root):
        monkeypatch.setenv("FOO", "bar")
        pdm(["run", "test_cmd"], obj=project)
        assert capfd.readouterr()[0].strip() == "1"

        result = pdm(["run", "test_cmd_no_expand"], obj=project)
        assert result.exit_code == 1

        pdm(["run", "test_script"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar"

        pdm(["run", "test_cmd_array"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar"

        pdm(["run", "test_shell"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar"


def test_run_expand_env_vars_from_config(project, pdm, capfd):
    (project.root / "test_script.py").write_text("import os; print(os.getenv('FOO'))")
    project.pyproject.settings["scripts"] = {
        "test_cmd": 'python -c "foo, bar = 0, 1;print(${FOO})"',
        "test_cmd_no_expand": "python -c 'print(${FOO})'",
        "test_script": "python test_script.py",
        "test_cmd_array": ["python", "test_script.py"],
        "test_shell": {"shell": "echo %FOO%" if os.name == "nt" else "echo $FOO"},
        "_": {"env": {"FOO": "bar"}},
    }
    project.pyproject.write()
    capfd.readouterr()
    with cd(project.root):
        pdm(["run", "test_cmd"], obj=project)
        assert capfd.readouterr()[0].strip() == "1"

        result = pdm(["run", "test_cmd_no_expand"], obj=project)
        assert result.exit_code == 1

        pdm(["run", "test_script"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar"

        pdm(["run", "test_cmd_array"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar"

        pdm(["run", "test_shell"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar"


def test_run_script_with_env_defined(project, pdm, capfd):
    (project.root / "test_script.py").write_text("import os; print(os.getenv('FOO'))")
    project.pyproject.settings["scripts"] = {"test_script": {"cmd": "python test_script.py", "env": {"FOO": "bar"}}}
    project.pyproject.write()
    capfd.readouterr()
    with cd(project.root):
        pdm(["run", "test_script"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar"


def test_run_script_with_dotenv_file(project, pdm, capfd, monkeypatch):
    (project.root / "test_script.py").write_text("import os; print(os.getenv('FOO'), os.getenv('BAR'))")
    project.pyproject.settings["scripts"] = {
        "test_override": {
            "cmd": "python test_script.py",
            "env_file": {"override": ".env"},
        },
        "test_default": {"cmd": "python test_script.py", "env_file": ".env"},
    }
    project.pyproject.write()
    monkeypatch.setenv("BAR", "foo")
    (project.root / ".env").write_text("FOO=bar\nBAR=override")
    capfd.readouterr()
    with cd(project.root):
        pdm(["run", "test_default"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar foo"
        pdm(["run", "test_override"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar override"


def test_run_script_override_global_env(project, pdm, capfd):
    (project.root / "test_script.py").write_text("import os; print(os.getenv('FOO'))")
    project.pyproject.settings["scripts"] = {
        "_": {"env": {"FOO": "bar"}},
        "test_env": {"cmd": "python test_script.py"},
        "test_env_override": {"cmd": "python test_script.py", "env": {"FOO": "foobar"}},
    }
    project.pyproject.write()
    capfd.readouterr()
    with cd(project.root):
        pdm(["run", "test_env"], obj=project)
        assert capfd.readouterr()[0].strip() == "bar"
        pdm(["run", "test_env_override"], obj=project)
        assert capfd.readouterr()[0].strip() == "foobar"


def test_run_show_list_of_scripts(project, pdm):
    project.pyproject.settings["scripts"] = {
        "test_composite": {"composite": ["test_cmd", "test_script", "test_shell"]},
        "test_cmd": "flask db upgrade",
        "test_multi": """\
            I am a multilines
            command
        """,
        "test_script": {"call": "test_script:main", "help": "call a python function"},
        "test_shell": {"shell": "echo $FOO", "help": "shell command"},
    }
    project.pyproject.write()
    result = pdm(["run", "--list"], obj=project)
    result_lines = result.output.splitlines()[3:]
    assert result_lines[0][1:-1].strip() == "test_cmd       │ cmd       │ flask db upgrade"
    sep = termui.Emoji.ARROW_SEPARATOR
    assert result_lines[1][1:-1].strip() == f"test_composite │ composite │ test_cmd {sep} test_script {sep} test_shell"
    assert result_lines[2][1:-1].strip() == f"test_multi     │ cmd       │ I am a multilines{termui.Emoji.ELLIPSIS}"
    assert result_lines[3][1:-1].strip() == "test_script    │ call      │ call a python function"
    assert result_lines[4][1:-1].strip() == "test_shell     │ shell     │ shell command"


def test_run_show_list_of_scripts_hide_internals(project, pdm):
    project.pyproject.settings["scripts"] = {
        "public": "true",
        "_internal": "true",
    }
    project.pyproject.write()
    result = pdm(["run", "--list"], obj=project)
    assert "public" in result.output
    assert "_internal" not in result.output


def test_run_json_list_of_scripts(project, pdm):
    project.pyproject.settings["scripts"] = {
        "_": {"env_file": ".env"},
        "test_composite": {"composite": ["test_cmd", "test_script", "test_shell"]},
        "test_cmd": "flask db upgrade",
        "test_multi": """\
            I am a multilines
            command
        """,
        "test_script": {"call": "test_script:main", "help": "call a python function"},
        "test_shell": {"shell": "echo $FOO", "help": "shell command"},
        "test_env": {"cmd": "true", "env": {"TEST": "value"}},
        "test_env_file": {"cmd": "true", "env_file": ".env"},
        "test_override": {"cmd": "true", "env_file": {"override": ".env"}},
        "test_site_packages": {"cmd": "true", "site_packages": True},
        "_private": "true",
    }
    project.pyproject.write()
    result = pdm(["run", "--json"], obj=project, strict=True)

    sep = termui.Emoji.ARROW_SEPARATOR
    assert json.loads(result.outputs) == {
        "_": {"name": "_", "help": "Shared options", "kind": "shared", "env_file": ".env"},
        "test_cmd": {"name": "test_cmd", "help": "flask db upgrade", "kind": "cmd", "args": "flask db upgrade"},
        "test_composite": {
            "name": "test_composite",
            "help": f"test_cmd {sep} test_script {sep} test_shell",
            "kind": "composite",
            "args": ["test_cmd", "test_script", "test_shell"],
        },
        "test_multi": {
            "name": "test_multi",
            "help": f"I am a multilines{termui.Emoji.ELLIPSIS}",
            "kind": "cmd",
            "args": "            I am a multilines\n            command\n        ",
        },
        "test_script": {
            "name": "test_script",
            "help": "call a python function",
            "kind": "call",
            "args": "test_script:main",
        },
        "test_shell": {"name": "test_shell", "help": "shell command", "kind": "shell", "args": "echo $FOO"},
        "test_env": {"name": "test_env", "help": "true", "kind": "cmd", "args": "true", "env": {"TEST": "value"}},
        "test_env_file": {"name": "test_env_file", "help": "true", "kind": "cmd", "args": "true", "env_file": ".env"},
        "test_override": {
            "name": "test_override",
            "help": "true",
            "kind": "cmd",
            "args": "true",
            "env_file.override": ".env",
        },
        "test_site_packages": {
            "name": "test_site_packages",
            "help": "true",
            "kind": "cmd",
            "args": "true",
            "site_packages": True,
        },
        "_private": {
            "name": "_private",
            "help": "true",
            "kind": "cmd",
            "args": "true",
        },
    }


@pytest.mark.usefixtures("local_finder")
@pytest.mark.parametrize("explicit_python", [True, False])
def test_run_with_another_project_root(project, pdm, capfd, explicit_python):
    project.pyproject.metadata["requires-python"] = ">=3.6"
    project.pyproject.write()
    pdm(["add", "first"], obj=project)
    with TemporaryDirectory(prefix="pytest-run-") as tmp_dir:
        Path(tmp_dir).joinpath("main.py").write_text("import first;print(first.first([0, False, 1, 2]))\n")
        capfd.readouterr()
        with cd(tmp_dir):
            args = ["run", "-p", str(project.root), "main.py"]
            if explicit_python:
                args.insert(len(args) - 1, "python")
            ret = pdm(args)
            out, err = capfd.readouterr()
            assert ret.exit_code == 0, err
            assert out.strip() == "1"


def test_import_another_sitecustomize(project, pdm, capfd):
    project.pyproject.metadata["requires-python"] = ">=2.7"
    project.pyproject.write()
    # a script for checking another sitecustomize is imported
    project.root.joinpath("foo.py").write_text("import os;print(os.getenv('FOO'))")
    # ensure there have at least one sitecustomize can be imported
    # there may have more than one sitecustomize.py in sys.path
    project.root.joinpath("sitecustomize.py").write_text("import os;os.environ['FOO'] = 'foo'")
    env = os.environ.copy()
    paths = env.get("PYTHONPATH")
    this_path = str(project.root)
    new_paths = [this_path] if not paths else [this_path, paths]
    env["PYTHONPATH"] = os.pathsep.join(new_paths)
    project._environment = None
    capfd.readouterr()
    with cd(project.root):
        result = pdm(["run", "python", "foo.py"], env=env)
    assert result.exit_code == 0, result.stderr
    out, _ = capfd.readouterr()
    assert out.strip() == "foo"


def test_run_with_patched_sysconfig(project, pdm, capfd):
    project.root.joinpath("script.py").write_text(
        """\
import sysconfig
import json
print(json.dumps(sysconfig.get_paths()))
"""
    )
    capfd.readouterr()
    with cd(project.root):
        result = pdm(["run", "python", "script.py"], obj=project)
    assert result.exit_code == 0
    out = json.loads(capfd.readouterr()[0])
    assert "__pypackages__" in out["purelib"]


def test_run_composite(project, pdm, capfd, _echo):
    project.pyproject.settings["scripts"] = {
        "first": "python echo.py First",
        "second": "python echo.py Second",
        "test": {"composite": ["first", "second"]},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "First CALLED" in out
    assert "Second CALLED" in out


def test_composite_stops_on_first_failure(project, pdm, capfd):
    project.pyproject.settings["scripts"] = {
        "first": {"cmd": ["python", "-c", "print('First CALLED')"]},
        "fail": "python -c 'raise Exception'",
        "second": "echo 'Second CALLED'",
        "test": {"composite": ["first", "fail", "second"]},
    }
    project.pyproject.write()
    capfd.readouterr()
    result = pdm(["run", "test"], obj=project)
    assert result.exit_code == 1
    out, _ = capfd.readouterr()
    assert "First CALLED" in out
    assert "Second CALLED" not in out


def test_composite_keep_going_on_failure(project, pdm, capfd):
    project.pyproject.settings["scripts"] = {
        "first": {"cmd": ["python", "-c", "print('First CALLED')"]},
        "fail": "python -c 'raise Exception'",
        "second": "echo 'Second CALLED'",
        "test": {"composite": ["first", "fail", "second"], "keep_going": True},
    }
    project.pyproject.write()
    capfd.readouterr()
    result = pdm(["run", "test"], obj=project)
    assert result.exit_code == 1
    out, _ = capfd.readouterr()
    assert "First CALLED" in out
    assert "Second CALLED" in out


def test_composite_inherit_env(project, pdm, capfd, _echo):
    project.pyproject.settings["scripts"] = {
        "first": {
            "cmd": "python echo.py First VAR",
            "env": {"VAR": "42"},
        },
        "second": {
            "cmd": "python echo.py Second VAR",
            "env": {"VAR": "42"},
        },
        "nested": {
            "composite": ["third"],
            "env": {"VAR": "42"},
        },
        "third": {
            "cmd": "python echo.py Third VAR",
            "env": {"VAR": "42"},
        },
        "test": {"composite": ["first", "second", "nested"], "env": {"VAR": "overridden"}},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "First CALLED with VAR=overridden" in out
    assert "Second CALLED with VAR=overridden" in out
    assert "Third CALLED with VAR=overridden" in out


def test_composite_fail_on_first_missing_task(project, pdm, capfd, _echo):
    project.pyproject.settings["scripts"] = {
        "first": "python echo.py First",
        "second": "python echo.py Second",
        "test": {"composite": ["first", "fail", "second"]},
    }
    project.pyproject.write()
    capfd.readouterr()
    result = pdm(["run", "test"], obj=project)
    assert result.exit_code == 1
    out, _ = capfd.readouterr()
    assert "First CALLED" in out
    assert "Second CALLED" not in out


def test_composite_fails_on_recursive_script(project, pdm):
    project.pyproject.settings["scripts"] = {
        "first": {"composite": ["first"]},
        "second": {"composite": ["third"]},
        "third": {"composite": ["second"]},
        "fourth": {"composite": ["python -V", "python -V"]},
        "fifth": {"composite": ["fourth", "fourth"]},
    }
    project.pyproject.write()
    result = pdm(["run", "first"], obj=project)
    assert result.exit_code == 1
    assert "Script first is recursive" in result.stderr

    result = pdm(["run", "second"], obj=project)
    assert result.exit_code == 1
    assert "Script second is recursive" in result.stderr

    result = pdm(["run", "fourth"], obj=project)
    assert result.exit_code == 0

    result = pdm(["run", "fifth"], obj=project)
    assert result.exit_code == 0


def test_composite_runs_all_hooks(project, pdm, capfd, _echo):
    project.pyproject.settings["scripts"] = {
        "test": {"composite": ["first", "second"]},
        "pre_test": "python echo.py Pre-Test",
        "post_test": "python echo.py Post-Test",
        "first": "python echo.py First",
        "pre_first": "python echo.py Pre-First",
        "second": "python echo.py Second",
        "post_second": "python echo.py Post-Second",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-Test CALLED" in out
    assert "Pre-First CALLED" in out
    assert "First CALLED" in out
    assert "Second CALLED" in out
    assert "Post-Second CALLED" in out
    assert "Post-Test CALLED" in out


def test_composite_pass_parameters_to_subtasks(project, pdm, capfd, _args):
    project.pyproject.settings["scripts"] = {
        "test": {"composite": ["first", "second"]},
        "pre_test": "python args.py Pre-Test",
        "post_test": "python args.py Post-Test",
        "first": "python args.py First",
        "pre_first": "python args.py Pre-First",
        "second": "python args.py Second",
        "post_second": "python args.py Post-Second",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test", "param=value"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-Test CALLED" in out
    assert "Pre-First CALLED" in out
    assert "First CALLED with param=value" in out
    assert "Second CALLED with param=value" in out
    assert "Post-Second CALLED" in out
    assert "Post-Test CALLED" in out


def test_composite_can_pass_parameters(project, pdm, capfd, _args):
    project.pyproject.settings["scripts"] = {
        "test": {"composite": ["first param=first", "second param=second"]},
        "pre_test": "python args.py Pre-Test",
        "post_test": "python args.py Post-Test",
        "first": "python args.py First",
        "pre_first": "python args.py Pre-First",
        "second": "python args.py Second",
        "post_second": "python args.py Post-Second",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-Test CALLED" in out
    assert "Pre-First CALLED" in out
    assert "First CALLED with param=first" in out
    assert "Second CALLED with param=second" in out
    assert "Post-Second CALLED" in out
    assert "Post-Test CALLED" in out


@pytest.mark.parametrize(
    "args,expected",
    (
        pytest.param(["-a"], "-a, ", id="with-args"),
        pytest.param([], "", id="without-args"),
    ),
)
def test_composite_only_pass_parameters_to_subtasks_with_args(project, pdm, capfd, _args, args, expected):
    project.pyproject.settings["scripts"] = {
        "test": {"composite": ["first", "second {args} key=value"]},
        "first": "python args.py First",
        "second": "python args.py Second",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "-v", "test", *args], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "First CALLED" in out
    assert f"Second CALLED with {expected}key=value" in out


@pytest.mark.parametrize(
    "args,expected",
    (
        pytest.param(["-a"], "-a", id="with-args"),
        pytest.param([], "--default", id="default"),
    ),
)
def test_composite_only_pass_parameters_to_subtasks_with_args_with_default(project, pdm, capfd, _args, args, expected):
    project.pyproject.settings["scripts"] = {
        "test": {"composite": ["first", "second {args:--default} key=value"]},
        "first": "python args.py First",
        "second": "python args.py Second",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "-v", "test", *args], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "First CALLED" in out
    assert f"Second CALLED with {expected}, key=value" in out


def test_composite_hooks_inherit_env(project, pdm, capfd, _echo):
    project.pyproject.settings["scripts"] = {
        "pre_task": {"cmd": "python echo.py Pre-Task VAR", "env": {"VAR": "42"}},
        "task": "python echo.py Task",
        "post_task": {"cmd": "python echo.py Post-Task VAR", "env": {"VAR": "42"}},
        "test": {"composite": ["task"], "env": {"VAR": "overridden"}},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-Task CALLED with VAR=overridden" in out
    assert "Task CALLED" in out
    assert "Post-Task CALLED with VAR=overridden" in out


def test_composite_inherit_env_in_cascade(project, pdm, capfd, _echo):
    project.pyproject.settings["scripts"] = {
        "_": {"env": {"FOO": "BAR", "TIK": "TOK"}},
        "pre_task": {
            "cmd": "python echo.py Pre-Task VAR FOO TIK",
            "env": {"VAR": "42", "FOO": "foobar"},
        },
        "task": {
            "cmd": "python echo.py Task VAR FOO TIK",
            "env": {"VAR": "42", "FOO": "foobar"},
        },
        "post_task": {
            "cmd": "python echo.py Post-Task VAR FOO TIK",
            "env": {"VAR": "42", "FOO": "foobar"},
        },
        "test": {"composite": ["task"], "env": {"VAR": "overridden"}},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-Task CALLED with VAR=overridden FOO=foobar TIK=TOK" in out
    assert "Task CALLED with VAR=overridden FOO=foobar TIK=TOK" in out
    assert "Post-Task CALLED with VAR=overridden FOO=foobar TIK=TOK" in out


def test_composite_inherit_dotfile(project, pdm, capfd, _echo):
    (project.root / ".env").write_text("VAR=42")
    (project.root / "override.env").write_text("VAR=overridden")
    project.pyproject.settings["scripts"] = {
        "pre_task": {"cmd": "python echo.py Pre-Task VAR", "env_file": ".env"},
        "task": {"cmd": "python echo.py Task VAR", "env_file": ".env"},
        "post_task": {"cmd": "python echo.py Post-Task VAR", "env_file": ".env"},
        "test": {"composite": ["task"], "env_file": "override.env"},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-Task CALLED with VAR=overridden" in out
    assert "Task CALLED with VAR=overridden" in out
    assert "Post-Task CALLED with VAR=overridden" in out


def test_resolve_env_vars_in_dotfile(project, pdm, capfd, _echo):
    (project.root / ".env").write_text("VAR=42\nFOO=${OUT}/${VAR}")
    project.pyproject.settings["scripts"] = {
        "_": {"env_file": ".env"},
        "test": {"cmd": "python echo.py Task FOO BAR", "env": {"BAR": "${FOO}/bar"}},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project, env={"OUT": "hello"})
    out, _ = capfd.readouterr()
    assert "Task CALLED with FOO=hello/42 BAR=hello/42/bar" in out


def test_composite_can_have_commands(project, pdm, capfd):
    project.pyproject.settings["scripts"] = {
        "task": {"cmd": ["python", "-c", 'print("Task CALLED")']},
        "test": {"composite": ["task", "python -c 'print(\"Command CALLED\")'"]},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "-v", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Task CALLED" in out
    assert "Command CALLED" in out


def test_run_shortcut(project, pdm, capfd):
    project.pyproject.settings["scripts"] = {
        "test": "echo 'Everything is fine'",
    }
    project.pyproject.write()
    capfd.readouterr()
    result = pdm(["test"], obj=project, strict=True)
    assert result.exit_code == 0
    out, _ = capfd.readouterr()
    assert "Everything is fine" in out


def test_run_shortcuts_dont_override_commands(project, pdm, capfd, mocker):
    do_lock = mocker.patch.object(actions, "do_lock")
    do_sync = mocker.patch.object(actions, "do_sync")
    project.pyproject.settings["scripts"] = {
        "install": "echo 'Should not run'",
    }
    project.pyproject.write()
    capfd.readouterr()
    result = pdm(["install"], obj=project, strict=True)
    assert result.exit_code == 0
    out, _ = capfd.readouterr()
    assert "Should not run" not in out
    do_lock.assert_called_once()
    do_sync.assert_called_once()


def test_run_shortcut_fail_with_usage_if_script_not_found(project, pdm):
    result = pdm(["whatever"], obj=project)
    assert result.exit_code != 0
    assert "Command not found: whatever" in result.stderr
    assert "usage" in result.stderr.lower()


@pytest.mark.parametrize(
    "args",
    [
        pytest.param(["-ko"], id="unknown param"),
        pytest.param(["pip", "--version"], id="not an user script"),
    ],
)
def test_empty_positional_args_still_display_usage(project, pdm, args):
    result = pdm(args, obj=project)
    assert result.exit_code != 0
    assert "usage" in result.stderr.lower()


def test_empty_positional_args_display_help(project, pdm):
    result = pdm([], obj=project)
    assert result.exit_code == 0
    assert "usage:" in result.output.lower()
    assert "commands:" in result.output.lower()
    assert "options:" in result.output.lower()


def test_run_script_changing_working_dir(project, pdm, capfd):
    project.root.joinpath("subdir").mkdir()
    project.root.joinpath("subdir", "file.text").write_text("Hello world\n")
    project.pyproject.settings["scripts"] = {
        "test_script": {"working_dir": "subdir", "cmd": "cat file.text"},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test_script"], obj=project, strict=True)
    assert capfd.readouterr()[0].strip() == "Hello world"


def test_run_script_with_inline_metadata(project, pdm, local_finder, local_finder_artifacts):
    with cd(project.root):
        project.root.joinpath("test_script.py").write_text(
            textwrap.dedent("""\
            from first import first

            assert first([0, False, 1, 2]) == 1
            """)
        )
        result = pdm(["run", "test_script.py"], obj=project)
        assert result.exit_code != 0

    local_artifacts_url = local_finder_artifacts.as_uri()

    project.root.joinpath("test_script.py").write_text(
        textwrap.dedent(f"""\
        # /// script
        # requires-python = ">=3.9"
        # dependencies = [
        #   "first",
        # ]
        #
        # [[tool.pdm.source]]
        # name = "pypi"
        # url = "{local_artifacts_url}"
        # type = "find_links"
        # ///
        from first import first

        assert first([0, False, 1, 2]) == 1
        """)
    )
    with cd(project.root):
        result = pdm(["run", "test_script.py"], obj=project)
        assert result.exit_code == 0


def test_run_script_pass_run_cwd(project, pdm, capfd):
    project.pyproject.settings["scripts"] = {
        "test_script": [
            "python",
            "-c",
            "import os;print(os.getenv('PDM_RUN_CWD'))",
        ]
    }
    project.pyproject.write()
    capfd.readouterr()
    result = pdm(["run", "test_script"], obj=project)
    assert result.exit_code == 0
    out, _ = capfd.readouterr()
    assert Path(out.strip()) == Path.cwd()


def test_run_script_pass_run_cwd_to_original_working_dir_when_working_dir_of_script_is_changed(project, pdm, capfd):
    project.root.joinpath("subdir").mkdir()
    project.root.joinpath("subdir", "test_script.py").write_text(
        textwrap.dedent("""\
           import os
           print(os.getenv('PDM_RUN_CWD', ''))
        """)
    )
    project.pyproject.settings["scripts"] = {
        "test_script": {"working_dir": "subdir", "cmd": "python test_script.py"},
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test_script"], obj=project, strict=True)
    assert capfd.readouterr()[0].strip() == str(Path.cwd())


def test_run_script_default_verbosity(project, pdm):
    project.pyproject.settings["scripts"] = {"test": {"cmd": "python -V", "help": "help"}}
    project.pyproject.write()

    result = pdm(["run", "test"], strict=True, obj=project)

    assert "task test" not in result.stderr
    assert "['python', '-V']" not in result.stderr
    assert "help" not in result.stderr


def test_run_script_default_verbosity_with_show_header(project, pdm):
    project.project_config.update({"scripts.show_header": True})
    project.pyproject.settings["scripts"] = {"test": {"cmd": "python -V", "help": "help"}}
    project.pyproject.write()

    result = pdm(["run", "test"], strict=True, obj=project)

    assert "task test" in result.stderr
    assert "['python', '-V']" not in result.stderr
    assert "python -V" not in result.stderr
    assert "help" in result.stderr


def test_run_script_default_verbosity_with_show_header_no_help(project, pdm):
    project.project_config.update({"scripts.show_header": True})
    project.pyproject.settings["scripts"] = {"test": {"cmd": "python -V"}}
    project.pyproject.write()

    result = pdm(["run", "test"], strict=True, obj=project)

    assert "task test" in result.stderr
    assert "python -V" in result.stderr
    assert "['python', '-V']" not in result.stderr


def test_run_script_verbose(project, pdm):
    project.pyproject.settings["scripts"] = {"test": {"cmd": "python -V", "help": "help"}}
    project.pyproject.write()

    result = pdm(["run", "-v", "test"], strict=True, obj=project)

    assert "task test" in result.stderr
    assert "['python', '-V']" in result.stderr
    assert "python -V" not in result.stderr
    assert "help" not in result.stderr


def test_run_composite_script_default_verbosity_with_show_header(project, pdm):
    project.project_config.update({"scripts.show_header": True})
    project.pyproject.settings["scripts"] = {
        "test": {"cmd": "python -V", "help": "help"},
        "parent": {"composite": ["test", "test"]},
    }
    project.pyproject.write()

    result = pdm(["run", "parent"], strict=True, obj=project)

    assert "task parent" in result.stderr
    assert f"test {termui.Emoji.ARROW_SEPARATOR} test" in result.stderr
    assert "['test', 'test']" not in result.stderr
    assert "task test" in result.stderr
    assert "python -V" not in result.stderr
    assert "['python', '-V']" not in result.stderr
    assert "help" in result.stderr


def test_run_composite_script_default_verbosity_with_show_header_and_help(project, pdm):
    project.project_config.update({"scripts.show_header": True})
    project.pyproject.settings["scripts"] = {
        "test": {"cmd": "python -V", "help": "help"},
        "parent": {"composite": ["test", "test"], "help": "composite"},
    }
    project.pyproject.write()

    result = pdm(["run", "parent"], strict=True, obj=project)

    assert "task parent" in result.stderr
    assert "composite" in result.stderr
    assert f"test {termui.Emoji.ARROW_SEPARATOR} test" not in result.stderr
    assert "['test', 'test']" not in result.stderr
    assert "task test" in result.stderr
    assert "['python', '-V']" not in result.stderr
    assert "python -V" not in result.stderr
    assert "help" in result.stderr


def test_run_composite_script_verbose(project, pdm):
    project.pyproject.settings["scripts"] = {
        "test": {"cmd": "python -V", "help": "help"},
        "parent": {"composite": ["test", "test"], "help": "composite"},
    }
    project.pyproject.write()

    result = pdm(["run", "-v", "parent"], strict=True, obj=project)

    assert "task parent" in result.stderr
    assert "composite" not in result.stderr
    assert f"test {termui.Emoji.ARROW_SEPARATOR} test" not in result.stderr
    assert "['test', 'test']" in result.stderr
    assert "task test" in result.stderr
    assert "['python', '-V']" in result.stderr
    assert "python -V" not in result.stderr
    assert "help" not in result.stderr



from __future__ import annotations

import shutil
import textwrap
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import httpx
import pytest
from pytest_mock import MockerFixture

from pdm.cli.commands.publish.package import PackageFile
from pdm.cli.commands.publish.repository import Repository
from pdm.models.auth import Keyring, keyring
from tests import FIXTURES


@pytest.fixture
def mock_run_gpg(mocker: MockerFixture):
    def mock_run_gpg(args):
        signature_file = args[-1] + ".asc"
        with open(signature_file, "wb") as f:
            f.write(b"fake signature")

    mocker.patch.object(PackageFile, "_run_gpg", side_effect=mock_run_gpg)


@pytest.fixture
def prepare_packages(tmp_path: Path):
    dist_path = tmp_path / "dist"
    dist_path.mkdir()
    for filename in [
        "demo-0.0.1-py2.py3-none-any.whl",
        "demo-0.0.1.tar.gz",
        "demo-0.0.1.zip",
    ]:
        shutil.copy2(FIXTURES / "artifacts" / filename, dist_path)


@pytest.fixture
def mock_pypi(mocker: MockerFixture):
    def send(request, **kwargs):
        # consume the data body to make the progress complete
        request.read()
        return httpx.Response(status_code=200, request=request)

    return mocker.patch("pdm.models.session.PDMPyPIClient.send", side_effect=send)


@pytest.fixture
def uploaded(mocker: MockerFixture):
    packages = []

    def fake_upload(package):
        packages.append(package)
        return httpx.Response(status_code=200, request=httpx.Request("POST", "https://upload.pypi.org/legacy/"))

    mocker.patch.object(Repository, "upload", side_effect=fake_upload)
    return packages


@dataclass
class PublishMock:
    mock_pypi: MagicMock
    uploaded: list[Any]


@pytest.fixture
# @pytest.mark.usefixtures("mock_run_gpg", "prepare_packages")
def mock_publish(mock_pypi, uploaded) -> PublishMock:
    return PublishMock(
        mock_pypi=mock_pypi,
        uploaded=uploaded,
    )


@pytest.fixture
def _echo(project):
    """
    Provides an echo.py script producing cross-platform expectable outputs
    """
    (project.root / "echo.py").write_text(
        textwrap.dedent(
            """\
            import os, sys, io
            sys.stdout = io.TextIOWrapper(sys.stdout.buffer, newline='\\n')
            name = sys.argv[1]
            vars = " ".join([f"{v}={os.getenv(v)}" for v in sys.argv[2:]])
            print(f"{name} CALLED with {vars}" if vars else f"{name} CALLED")
            """
        )
    )


@pytest.fixture(name="keyring")
def keyring_fixture(mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch) -> Keyring:
    from unearth.auth import AuthInfo, KeyringBaseProvider

    class MockKeyringProvider(KeyringBaseProvider):
        def __init__(self) -> None:
            self._store: dict[str, dict[str, str]] = {}

        def save_auth_info(self, url: str, username: str, password: str) -> None:
            self._store.setdefault(url, {})[username] = password

        def get_auth_info(self, url: str, username: str | None) -> AuthInfo | None:
            d = self._store.get(url, {})
            if username is not None and username in d:
                return username, d[username]
            if username is None and d:
                return next(iter(d.items()))
            return None

        def delete_auth_info(self, url: str, username: str) -> None:
            self._store.get(url, {}).pop(username, None)

    provider = MockKeyringProvider()
    mocker.patch("unearth.auth.get_keyring_provider", return_value=provider)
    monkeypatch.setattr(keyring, "provider", provider)
    monkeypatch.setattr(keyring, "enabled", True)
    keyring.get_auth_info.cache_clear()
    return keyring



import logging
import textwrap

import pytest

from pdm.exceptions import ProjectError
from pdm.formats import MetaConvertError
from pdm.models.setup import Setup


def test_setup_update_truthiness_semantics():
    base = Setup(name="foo", install_requires=["a"], summary=None)
    other = Setup(name=None, install_requires=[], summary="some desc")
    base.update(other)
    assert base.name == "foo"  # not overridden by falsy
    assert base.install_requires == ["a"]  # not overridden by empty list
    assert base.summary == "some desc"  # overridden by truthy


def test_parse_setup_py_with_kwargs_dict_and_variables(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup
        reqs = ['click', 'requests']
        extras_list = ['rich']
        extras = {'tui': extras_list}
        kwargs = dict(name='foo', version='0.1.0', install_requires=reqs,
                      python_requires='>=3.8', extras_require=extras)
        setup(**kwargs)
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    assert Setup.from_directory(tmp_path) == Setup(
        name="foo",
        version="0.1.0",
        install_requires=["click", "requests"],
        extras_require={"tui": ["rich"]},
        python_requires=">=3.8",
    )


def test_setup_as_dict():
    s = Setup(
        name="n", version="1", install_requires=["a"], extras_require={"x": ["b"]}, python_requires=">=3.8", summary="d"
    )
    d = s.as_dict()
    assert d["name"] == "n"
    assert d["version"] == "1"
    assert d["install_requires"] == ["a"]
    assert d["extras_require"] == {"x": ["b"]}
    assert d["python_requires"] == ">=3.8"
    assert d["summary"] == "d"


@pytest.mark.parametrize(
    "content",
    [
        # Last element is an if but not a Compare
        """
if True:
    pass
""",
        # Compare but left is Attribute not Name
        """
import pkg
if pkg.__name__ == "__main__":
    pass
""",
        # Compare left is Name but not __name__
        """
name = "x"
if name == "__main__":
    pass
""",
    ],
)
def test_parse_setup_py_no_setup_call_branches(tmp_path, content):
    tmp_path.joinpath("setup.py").write_text(textwrap.dedent(content))
    assert Setup.from_directory(tmp_path) == Setup()


def test_parse_setup_py_irrelevant_call_and_assignment(tmp_path):
    content = textwrap.dedent(
        """
        def foo():
            return 1
        x = 1
        foo()
        # no setup() call here
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    assert Setup.from_directory(tmp_path) == Setup()


@pytest.mark.parametrize(
    "content, expected",
    [
        # No sections -> defaults (version defaults to "0.0.0")
        ("", Setup(version="0.0.0")),
        # Empty metadata section only
        ("[metadata]\n", Setup(version="0.0.0")),
        # Empty options section only
        ("[options]\n", Setup(version="0.0.0")),
        # Name only, version still defaults to 0.0.0
        ("[metadata]\nname = foo\n", Setup(name="foo", version="0.0.0")),
        # Version attr -> keep default 0.0.0
        ("[metadata]\nversion = attr:foo.__version__\n", Setup(version="0.0.0")),
    ],
)
def test_parse_setup_cfg_missing_sections_and_options(tmp_path, content, expected):
    tmp_path.joinpath("setup.cfg").write_text(content)
    assert Setup.from_directory(tmp_path) == expected


def test_find_setup_call_in_function_and_if_main(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup

        def inner():
            setup(name="foo", version="0.1.0")

        if __name__ == "__main__":
            inner()
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    assert Setup.from_directory(tmp_path) == Setup(name="foo", version="0.1.0")


def test_if_main_direct_setup_hits_concat_body_return(tmp_path):
    content = textwrap.dedent(
        """
        if __name__ == "__main__":
            from setuptools import setup
            setup(name="foo2", version="0.3.0")
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    assert Setup.from_directory(tmp_path) == Setup(name="foo2", version="0.3.0")


def test_from_directory_precedence_and_falsy_update(tmp_path):
    pyproject = textwrap.dedent(
        """
        [project]
        name = "name_py"
        version = "0.1.0"
        requires-python = ">=3.7"
        dependencies = ["a"]
        [project.optional-dependencies]
        tui = ["r1"]
        """
    )
    tmp_path.joinpath("pyproject.toml").write_text(pyproject)

    setup_cfg = textwrap.dedent(
        """
        [metadata]
        name = name_cfg
        version = 0.1.1

        [options]
        python_requires = >=3.8
        install_requires =
            b

        [options.extras_require]
        tui =
            r2
        """
    )
    tmp_path.joinpath("setup.cfg").write_text(setup_cfg)

    setup_py = textwrap.dedent(
        """
        from setuptools import setup
        setup(name='name_py', version='0.2.0', install_requires=[], extras_require={'tui': ['r3']})
        """
    )
    tmp_path.joinpath("setup.py").write_text(setup_py)

    result = Setup.from_directory(tmp_path)
    assert result.name == "name_py"  # from setup.py (last wins)
    assert result.version == "0.2.0"  # from setup.py
    assert result.install_requires == ["b"]  # not overridden by empty list in setup.py
    assert result.extras_require == {"tui": ["r3"]}  # overridden by setup.py (truthy)
    assert result.python_requires == ">=3.8"  # from setup.cfg (later non-empty vs pyproject)


def test_read_pyproject_toml_project_error_returns_empty(tmp_path, mocker):
    # Create a dummy PyProject that raises ProjectError on unwrap()
    import pdm.project.project_file as project_file

    mocker.patch.object(project_file.PyProject, "_convert_pyproject", side_effect=ProjectError("boom"))

    tmp_path.joinpath("pyproject.toml").write_text("[project]\nname='x'")
    assert Setup.from_directory(tmp_path) == Setup()  # empty due to ProjectError


def test_read_pyproject_toml_metaconverter_error_uses_partial_data_and_logs(tmp_path, mocker, caplog):
    partial = {
        "name": "foo",
        "version": "0.1.0",
        "description": "desc",
        "dependencies": ["a"],
        "optional-dependencies": {"tui": ["r"]},
        "requires-python": ">=3.8",
    }

    import pdm.project.project_file as project_file

    mocker.patch.object(
        project_file.PyProject, "_convert_pyproject", side_effect=MetaConvertError(["e1"], data=partial, settings={})
    )

    caplog.set_level("WARNING")
    logging.getLogger("pdm.termui").addHandler(caplog.handler)
    tmp_path.joinpath("pyproject.toml").write_text("[tool.other]\nfoo='bar'")
    result = Setup.from_directory(tmp_path)

    # Check fields are populated from partial data
    assert result.name == "foo"
    assert result.version == "0.1.0"
    assert result.summary == "desc"
    assert result.install_requires == ["a"]
    assert result.extras_require == {"tui": ["r"]}
    assert result.python_requires == ">=3.8"

    # Check a warning was logged
    assert any("Error parsing pyproject.toml" in message for message in caplog.messages)


def test_setup_distribution_metadata_and_requires_markers():
    dist = Setup(
        name="pack",
        version="1.0.0",
        summary="desc",
        python_requires=">=3.8",
        install_requires=["base>=1"],
        extras_require={
            "tui": [
                "rich",
                'foo; python_version >= "3.9"',
                'bar; python_version < "3.8" or sys_platform == "win32"',
            ]
        },
    ).as_dist()

    meta = dist.metadata
    assert meta == {
        "Name": "pack",
        "Version": "1.0.0",
        "Summary": "desc",
        "Requires-Python": ">=3.8",
    }

    reqs = dist.requires
    # Contains base
    assert any(r.startswith("base>=1") for r in reqs)
    # Extra-only requirement with no existing marker
    assert any(r.startswith("rich") and 'extra == "tui"' in r for r in reqs)
    # Existing marker combined without parentheses
    assert any(r.startswith("foo") and 'python_version >= "3.9" and extra == "tui"' in r for r in reqs)
    # Existing marker with OR and repeated extra markers (no parentheses)
    assert any(
        r.startswith("bar")
        and 'python_version < "3.8" and extra == "tui" or sys_platform == "win32" and extra == "tui"' in r
        for r in reqs
    )


def test_read_text_and_locate_file_smoke():
    dist = Setup(name="x", version="1").as_dist()
    assert dist.read_text("any") is None
    p = dist.locate_file("whatever")
    from pathlib import Path as _P

    assert isinstance(p, _P)


def test_find_setup_call_skips_non_call_expr_then_finds_setup(tmp_path):
    content = textwrap.dedent(
        """
        "module docstring"
        from setuptools import setup
        setup(name='d', version='0.1.0')
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    assert Setup.from_directory(tmp_path) == Setup(name="d", version="0.1.0")


def test_find_setup_call_skips_non_functiondef_element_then_finds_setup(tmp_path):
    content = textwrap.dedent(
        """
        class X: pass
        from setuptools import setup
        setup(name='e', version='0.1.0')
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    assert Setup.from_directory(tmp_path) == Setup(name="e", version="0.1.0")


def test_find_setup_call_skips_non_setup_call_then_finds_setup(tmp_path):
    content = textwrap.dedent(
        """
        def foo(): pass
        foo()
        from setuptools import setup
        setup(name='f', version='0.1.0')
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    assert Setup.from_directory(tmp_path) == Setup(name="f", version="0.1.0")


def test_install_requires_name_resolves_to_non_list(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup
        reqs = 'notalist'
        setup(name='g', version='0.1.0', install_requires=reqs)
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    result = Setup.from_directory(tmp_path)
    assert result.name == "g"
    assert result.install_requires == []


def test_extras_require_name_is_undefined_returns_empty(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup
        setup(name='h', version='0.1.0', extras_require=extras)
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    result = Setup.from_directory(tmp_path)
    assert result.extras_require == {}


def test_extras_require_name_resolves_to_non_dict(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup
        extras = ['x']
        setup(name='i', version='0.1.0', extras_require=extras)
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    result = Setup.from_directory(tmp_path)
    assert result.extras_require == {}


def test_single_string_name_variable_not_string(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup
        namev = 123
        setup(name=namev, version='0.1.0')
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    result = Setup.from_directory(tmp_path)
    assert result.name is None


def test_extras_require_from_kwargs_non_dict_non_call_returns_empty(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup
        kwargs = "notadict"
        setup(**kwargs)
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    result = Setup.from_directory(tmp_path)
    assert result.extras_require == {}
    assert result.version == "0.0.0"


def test_extras_require_from_kwargs_call_func_is_attribute_returns_empty(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup
        class Obj:
            def build(self):
                return {}
        obj = Obj()
        kwargs = obj.build()
        setup(**kwargs)
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    result = Setup.from_directory(tmp_path)
    assert result.extras_require == {}


def test_extras_require_from_kwargs_call_func_name_not_dict_returns_empty(tmp_path):
    content = textwrap.dedent(
        """
        from setuptools import setup
        def make_kwargs(**kw):
            return kw
        kwargs = make_kwargs(name='n', version='0.1.0')
        setup(**kwargs)
        """
    )
    tmp_path.joinpath("setup.py").write_text(content)
    result = Setup.from_directory(tmp_path)
    # kwargs created by a function call that is not dict() won't be introspected
    assert result.name is None
    assert result.version == "0.0.0"
    assert result.extras_require == {}



import shutil
from pathlib import Path

import pytest

from pdm.models.backends import _BACKENDS, get_backend, get_relative_path
from pdm.project import Project
from pdm.utils import cd
from tests import FIXTURES


def _setup_backend(project: Project, backend: str):
    project.pyproject.metadata["requires-python"] = ">=3.6"
    backend_cls = get_backend(backend)
    project.pyproject.build_system.update(backend_cls.build_system())
    project.root.joinpath("test_project").mkdir()
    project.root.joinpath("test_project/__init__.py").touch()
    if backend == "setuptools":
        project.pyproject._data.setdefault("tool", {}).setdefault("setuptools", {}).update(packages=["test_project"])
    elif backend == "hatchling":
        project.pyproject._data.setdefault("tool", {}).setdefault("hatch", {}).setdefault("metadata", {}).update(
            {"allow-direct-references": True}
        )
    project.pyproject.write()
    project._environment = None
    assert isinstance(project.backend, backend_cls)


@pytest.mark.parametrize("backend", _BACKENDS.keys())
def test_project_backend(project, working_set, backend, pdm):
    _setup_backend(project, backend)
    shutil.copytree(FIXTURES / "projects/demo", project.root / "demo")
    project.root.joinpath("sub").mkdir()
    with cd(project.root.joinpath("sub")):
        pdm(["add", "--no-self", "../demo"], obj=project, strict=True)
        assert "idna" in working_set
        assert "demo" in working_set
        dep = project.pyproject.metadata["dependencies"][0]
        demo_path = project.root.joinpath("demo")
        demo_url = demo_path.as_uri()
        if backend == "pdm-backend":
            assert dep == "demo @ file:///${PROJECT_ROOT}/demo"
        elif backend == "hatchling":
            assert dep == "demo @ {root:uri}/demo"
        else:
            assert dep == f"demo @ {demo_url}"
        assert project.backend.expand_line(dep) == f"demo @ {demo_url}"
        if backend not in ("hatchling", "pdm-backend"):
            candidate = project.make_self_candidate()
            # We skip hatchling here to avoid installing hatchling into the build env
            metadata_dependency = candidate.prepare(project.environment).metadata.requires[0]
            assert metadata_dependency == f"demo @ {demo_url}"


def test_hatch_expand_variables(monkeypatch):
    root = Path().absolute()
    root_url = root.as_uri()
    backend = get_backend("hatchling")(root)
    monkeypatch.setenv("BAR", "bar")
    assert backend.expand_line("demo @ {root:uri}/demo") == f"demo @ {root_url}/demo"
    assert backend.expand_line("demo=={env:FOO:{env:BAR}}") == "demo==bar"
    assert backend.relative_path_to_url("demo package") == "{root:uri}/demo%20package"
    assert backend.relative_path_to_url("../demo") == "{root:uri}/../demo"


def test_pdm_backend_expand_variables(monkeypatch):
    root = Path().absolute()
    root_url = root.as_uri()
    backend = get_backend("pdm-backend")(root)
    monkeypatch.setenv("BAR", "bar")
    assert backend.expand_line("demo @ file:///${PROJECT_ROOT}/demo") == f"demo @ {root_url}/demo"
    assert backend.expand_line("demo==${BAR}") == "demo==bar"
    assert backend.relative_path_to_url("demo package") == "file:///${PROJECT_ROOT}/demo%20package"
    assert backend.relative_path_to_url("../demo") == "file:///${PROJECT_ROOT}/../demo"


@pytest.mark.parametrize(
    "url,path",
    [
        ("file:///foo/bar", None),
        ("https://example.org", None),
        ("file:///${PROJECT_ROOT}/demo%20package", "demo package"),
        ("file:///${PROJECT_ROOT}/../demo", "../demo"),
        ("{root:uri}/demo%20package", "demo package"),
    ],
)
def test_get_relative_path(url, path):
    assert get_relative_path(url) == path



def test_help_with_unknown_arguments(pdm):
    result = pdm(["add", "--unknown-args"])
    assert "usage: pdm add " in result.stderr.lower()
    assert result.exit_code == 2


def test_output_similar_command_when_typo(pdm):
    result = pdm(["instal"])
    assert "install" in result.stderr
    assert result.exit_code == 2



import sys
from unittest import mock

import pytest

from pdm.cli.commands.base import BaseCommand
from pdm.compat import importlib_metadata
from pdm.project.config import ConfigItem
from pdm.utils import cd


class HelloCommand(BaseCommand):
    def add_arguments(self, parser) -> None:
        parser.add_argument("-n", "--name", help="The person's name")

    def handle(self, project, options) -> None:
        greeting = "Hello world"
        if options.name:
            greeting = f"Hello, {options.name}"
        print(greeting)


def new_command(core):
    core.register_command(HelloCommand, "hello")


def replace_command(core):
    core.register_command(HelloCommand, "info")


def add_new_config(core):
    core.add_config("foo", ConfigItem("Test config", "bar"))


def make_entry_point(plugin):
    ret = mock.Mock()
    ret.load.return_value = plugin
    return ret


def test_plugin_new_command(pdm, mocker, project, core):
    mocker.patch.object(
        importlib_metadata,
        "entry_points",
        return_value=[make_entry_point(new_command)],
    )
    core.init_parser()
    core.load_plugins()
    result = pdm(["--help"], obj=project)
    assert "hello" in result.output

    result = pdm(["hello"], obj=project)
    assert result.output.strip() == "Hello world"

    result = pdm(["hello", "-n", "Frost"], obj=project)
    assert result.output.strip() == "Hello, Frost"


def test_plugin_replace_command(pdm, mocker, project, core):
    mocker.patch.object(
        importlib_metadata,
        "entry_points",
        return_value=[make_entry_point(replace_command)],
    )
    core.init_parser()
    core.load_plugins()

    result = pdm(["info"], obj=project)
    assert result.output.strip() == "Hello world"

    result = pdm(["info", "-n", "Frost"], obj=project)
    assert result.output.strip() == "Hello, Frost"


def test_load_multiple_plugins(pdm, mocker, core):
    mocker.patch.object(
        importlib_metadata,
        "entry_points",
        return_value=[make_entry_point(new_command), make_entry_point(add_new_config)],
    )
    core.init_parser()
    core.load_plugins()

    result = pdm(["hello"])
    assert result.output.strip() == "Hello world", result.outputs

    result = pdm(["config", "foo"])
    assert result.output.strip() == "bar"


def test_old_entry_point_compatibility(pdm, mocker, core):
    def get_entry_points(group):
        if group == "pdm":
            return [make_entry_point(new_command)]
        if group == "pdm.plugin":
            return [make_entry_point(add_new_config)]
        return []

    mocker.patch.object(importlib_metadata, "entry_points", side_effect=get_entry_points)
    core.init_parser()
    core.load_plugins()

    result = pdm(["hello"])
    assert result.output.strip() == "Hello world"

    result = pdm(["config", "foo"])
    assert result.output.strip() == "bar"


@pytest.mark.usefixtures("local_finder")
def test_project_plugin_library(pdm, project, core, monkeypatch):
    monkeypatch.setattr(sys, "path", sys.path[:])
    project.pyproject.settings["plugins"] = ["pdm-hello"]
    pdm(["install", "--plugins"], obj=project, strict=True)
    assert project.root.joinpath(".pdm-plugins").exists()
    assert "pdm-hello" not in project.environment.get_working_set()
    with cd(project.root):
        core.load_plugins()
        result = pdm(["hello", "Frost"], strict=True)
    assert result.stdout.strip() == "Hello, Frost!"


@pytest.mark.parametrize(
    "req_str",
    [
        "-e file:///${PROJECT_ROOT}/plugins/test-plugin-pdm",
        "-e test_plugin_pdm @ file:///${PROJECT_ROOT}/plugins/test-plugin-pdm",
    ],
)
@pytest.mark.usefixtures("local_finder")
def test_install_local_plugin_without_name(pdm, project, core, req_str):
    import shutil

    from . import FIXTURES

    test_plugin_path = FIXTURES / "projects" / "test-plugin-pdm"
    project.root.joinpath("plugins").mkdir(exist_ok=True)
    shutil.copytree(test_plugin_path, project.root / "plugins" / "test-plugin-pdm", dirs_exist_ok=True)

    project.pyproject.settings["plugins"] = [req_str]
    project.pyproject.write()
    with cd(project.root):
        result = pdm(["install", "--plugins", "-vv"], obj=project, strict=True)

        assert project.root.joinpath(".pdm-plugins").exists()
        core.load_plugins()
        result = pdm(["hello", "--name", "Frost"], strict=True)
    assert result.stdout.strip() == "Hello, Frost"



import pathlib
import sys
import unittest.mock as mock
from datetime import datetime, timezone
from pathlib import Path

import pytest
import tomlkit

from pdm import utils
from pdm._types import RepositoryConfig
from pdm.cli import utils as cli_utils
from pdm.cli.filters import GroupSelection
from pdm.exceptions import PdmException, PdmUsageError, PDMWarning


@pytest.mark.parametrize(
    "given, dirname",
    [
        ((None, None, None), "test_dirname1"),
        (("test_suffix", None, None), "test_dirname2"),
        (("test_suffix", "test_prefix", None), "test_dirname3"),
        (("test_suffix", "test_prefix", "test_dir"), "test_dirname4"),
        ((None, "test_prefix", None), "test_dirname5"),
        ((None, "test_prefix", "test_dir"), "test_dirname6"),
        ((None, None, "test_dir"), "test_dirname7"),
        ((None, "test_prefix", "test_dir"), "test_dirname8"),
        (("test_prefix", None, "test_dir"), "test_dirname9"),
    ],
)
@mock.patch("pdm.utils.atexit.register")
@mock.patch("pdm.utils.os.makedirs")
@mock.patch("pdm.utils.tempfile.mkdtemp")
def test_create_tracked_tempdir(mock_tempfile_mkdtemp, mock_os_makedirs, mock_atexit_register, given, dirname):
    test_suffix, test_prefix, _ = given
    mock_tempfile_mkdtemp.return_value = dirname
    received_dirname = utils.create_tracked_tempdir(suffix=test_suffix, prefix=test_prefix, dir=dirname)
    mock_tempfile_mkdtemp.assert_called_once_with(suffix=test_suffix, prefix=test_prefix, dir=dirname)
    mock_os_makedirs.assert_called_once_with(dirname, mode=0o777, exist_ok=True)
    mock_atexit_register.assert_called()
    assert received_dirname == dirname


def test_get_trusted_hosts(mocker):
    repository_configs = [
        {
            "url": "https://pypi.org",
            "verify_ssl": False,
        },
        {"url": "https://untrusted.pypi.index", "verify_ssl": True},
        {"url": "https://user:password@trusted1.pypi.index", "verify_ssl": False},
        {"url": "https://user:password@trusted2.pypi.index", "verify_ssl": False},
    ]
    sources = [mocker.create_autospec(RepositoryConfig, instance=False, **config) for config in repository_configs]
    expected = [
        "pypi.org",
        "trusted1.pypi.index",
        "trusted2.pypi.index",
    ]
    received = utils.get_trusted_hosts(sources)
    assert received == expected


@pytest.mark.parametrize(
    "given, expected",
    [
        ("scheme://netloc", "scheme://netloc"),
        ("scheme://netloc/path", "scheme://netloc/path"),
        ("scheme://netloc/path/#", "scheme://netloc/path/"),
        ("scheme://netloc/path#fragment", "scheme://netloc/path"),
        ("scheme://netloc/path;parameters?query#fragment", "scheme://netloc/path;parameters?query"),
    ],
)
def test_url_without_fragments(given, expected):
    received = utils.url_without_fragments(given)
    assert received == expected


@pytest.mark.parametrize(
    "given, expected",
    [
        ((["abc", "def", "ghi"], "/"), ["abc", "/", "def", "/", "ghi"]),
        (([], "/"), []),
        ((["abc"], "/"), ["abc"]),
    ],
)
def test_join_list_with(given, expected):
    items, sep = given
    received = utils.join_list_with(items, sep)
    assert received == expected


class TestGetUserEmailFromGit:
    @mock.patch("pdm.utils.shutil.which", return_value=None)
    def test_no_git(self, no_git_patch):
        with no_git_patch:
            assert utils.get_user_email_from_git() == ("", "")

    @mock.patch(
        "pdm.utils.subprocess.check_output",
        side_effect=[
            utils.subprocess.CalledProcessError(-1, ["git", "config", "user.name"], "No username"),
            utils.subprocess.CalledProcessError(-1, ["git", "config", "user.email"], "No email"),
        ],
    )
    @mock.patch("pdm.utils.shutil.which", return_value="git")
    def test_no_git_username_and_email(self, git_patch, no_git_username_and_email_patch):
        with git_patch:
            with no_git_username_and_email_patch:
                assert utils.get_user_email_from_git() == ("", "")

    @mock.patch(
        "pdm.utils.subprocess.check_output",
        side_effect=[
            "username",
            utils.subprocess.CalledProcessError(-1, ["git", "config", "user.email"], "No email"),
        ],
    )
    @mock.patch("pdm.utils.shutil.which", return_value="git")
    def test_no_git_email(self, git_patch, no_git_email_patch):
        with git_patch:
            with no_git_email_patch:
                assert utils.get_user_email_from_git() == ("username", "")

    @mock.patch(
        "pdm.utils.subprocess.check_output",
        side_effect=[utils.subprocess.CalledProcessError(-1, ["git", "config", "user.name"], "No username"), "email"],
    )
    @mock.patch("pdm.utils.shutil.which", return_value="git")
    def test_no_git_username(self, git_patch, no_git_username_patch):
        with git_patch:
            with no_git_username_patch:
                assert utils.get_user_email_from_git() == ("", "email")

    @mock.patch("pdm.utils.subprocess.check_output", side_effect=["username", "email"])
    @mock.patch("pdm.utils.shutil.which", return_value="git")
    def test_git_username_and_email(self, git_patch, git_username_and_email_patch):
        with git_patch:
            with git_username_and_email_patch:
                assert utils.get_user_email_from_git() == ("username", "email")


@pytest.mark.parametrize(
    "given,expected",
    [
        ("git@github.com/pdm-project/pdm", "ssh://git@github.com/pdm-project/pdm"),
        ("ssh://git@github.com/pdm-project/pdm", "ssh://git@github.com/pdm-project/pdm"),
        ("git+ssh://git@github.com/pdm-project/pdm", "git+ssh://git@github.com/pdm-project/pdm"),
        ("https://git@github.com/pdm-project/pdm", "https://git@github.com/pdm-project/pdm"),
        ("file:///my/local/pdm-project/pdm", "file:///my/local/pdm-project/pdm"),
    ],
)
def test_add_ssh_scheme_to_git_uri(given, expected):
    assert utils.add_ssh_scheme_to_git_uri(given) == expected


class TestUrlToPath:
    def test_non_file_url(self):
        with pytest.raises(ValueError):
            utils.url_to_path("not_a_file_scheme://netloc/path")

    @pytest.mark.skipif(sys.platform.startswith("win"), reason="Non-Windows test")
    def test_non_windows_non_local_file_url(self):
        with pytest.raises(ValueError):
            utils.url_to_path("file://non_local_netloc/file/url")

    @pytest.mark.skipif(sys.platform.startswith("win"), reason="Non-Windows test")
    def test_non_windows_localhost_local_file_url(self):
        assert utils.url_to_path("file://localhost/local/file/path") == "/local/file/path"

    @pytest.mark.skipif(not sys.platform.startswith("win"), reason="Windows test")
    def test_windows_localhost_local_file_url(self):
        assert utils.url_to_path("file://localhost/local/file/path") == "\\local\\file\\path"


@pytest.mark.parametrize(
    "given,expected",
    [
        ("test", "test"),
        ("", ""),
        ("${FOO}", "hello"),
        ("$FOO", "$FOO"),
        ("${BAR}", ""),
        ("%FOO%", "%FOO%"),
        ("${FOO}_${FOO}", "hello_hello"),
    ],
)
def test_expand_env_vars(given, expected, monkeypatch):
    monkeypatch.setenv("FOO", "hello")
    assert utils.expand_env_vars(given) == expected


@pytest.mark.parametrize(
    "given,expected",
    [
        ("https://example.org/path?arg=1", "https://example.org/path?arg=1"),
        (
            "https://${FOO}@example.org/path?arg=1",
            "https://token%3Aoidc%2F1@example.org/path?arg=1",
        ),
        (
            "https://${FOO}:${BAR}@example.org/path?arg=1",
            "https://token%3Aoidc%2F1:p%40ssword@example.org/path?arg=1",
        ),
        (
            "https://${FOOBAR}@example.org/path?arg=1",
            "https://@example.org/path?arg=1",
        ),
    ],
)
def test_expand_env_vars_in_auth(given, expected, monkeypatch):
    monkeypatch.setenv("FOO", "token:oidc/1")
    monkeypatch.setenv("BAR", "p@ssword")
    assert utils.expand_env_vars_in_auth(given) == expected


@pytest.mark.parametrize(
    "os_name,given,expected",
    [
        ("posix", ("match", "repl", "/a/b/match/c/match/d/e"), "/a/b/repl/c/repl/d/e"),
        ("posix", ("old", "new", "/path/to/old/pdm"), "/path/to/new/pdm"),
        ("posix", ("match", "repl", "match/a/math/b/match/c"), "repl/a/math/b/repl/c"),
        ("posix", ("match", "repl", "/some/path"), "/some/path"),
        ("posix", ("match", "repl", ""), ""),
        ("nt", ("old", "new", "C:\\Path\\tO\\old\\pdm"), "C:/Path/tO/new/pdm"),
        ("nt", ("old", "new", "C:\\Path\\tO\\Old\\pdm"), "C:/Path/tO/new/pdm"),
        ("nt", ("old", "new", "C:\\no\\matching\\path"), "C:/no/matching/path"),
    ],
)
def test_path_replace(os_name, given, expected):
    with mock.patch("pdm.utils.os_name", os_name):
        pattern, replace_with, dest = given
        assert utils.path_replace(pattern, replace_with, dest) == expected


# Only testing POSIX-style paths here
@pytest.mark.parametrize(
    "given,expected",
    [
        (("/", "/"), True),
        (("/a", "/"), True),
        (("/a/b", "/a"), True),
        (("/a", "/b"), False),
        (("a", "b"), False),
        (("/a/b", "/c/d"), False),
        (("/a/b/c", "/a"), True),
        (("../a/b/c", "../a"), True),
    ],
)
def test_is_path_relative_to(given, expected):
    path, other = given
    assert utils.is_path_relative_to(path, other) == expected


class TestGetVenvLikePrefix:
    def test_conda_env_with_conda_meta_in_bin(self, tmp_path: Path):
        path = tmp_path / "conda/bin/python3"
        path_parent = path.parent
        path_parent.mkdir(parents=True)
        path.touch()
        path_parent.joinpath("conda-meta").mkdir()
        received = utils.get_venv_like_prefix(path)
        expected = path_parent, True
        assert received == expected

    def test_py_env_with_pyvenv_cfg(self, tmp_path: Path):
        path = tmp_path / "venv/bin/python3"
        bin_path = path.parent
        venv_path = path.parent.parent
        bin_path.mkdir(parents=True)
        path.touch()
        venv_path.joinpath("pyvenv.cfg").touch()

        received = utils.get_venv_like_prefix(str(path))
        expected = venv_path, False
        assert received == expected

    def test_conda_env_with_conda_meta(self, tmp_path: Path):
        path = tmp_path / "conda/bin/python3"
        interpreter_bin_path = path.parent
        interpreter_bin_parent_path = interpreter_bin_path.parent
        interpreter_bin_path.mkdir(parents=True)
        path.touch()
        interpreter_bin_parent_path.joinpath("conda-meta").mkdir()

        received = utils.get_venv_like_prefix(str(path))
        expected = interpreter_bin_parent_path, True
        assert received == expected

    def test_virtual_env(self, monkeypatch):
        path = Path("/my/venv")
        expected = path, False
        monkeypatch.setenv("VIRTUAL_ENV", str(path))
        received = utils.get_venv_like_prefix(path.joinpath("bin", "python3"))
        assert received == expected

    def test_conda_virtual_env(self, monkeypatch):
        path = Path("/my/conda/venv")
        expected = path, True
        monkeypatch.setenv("CONDA_PREFIX", str(path))
        received = utils.get_venv_like_prefix(path.joinpath("bin", "python3"))
        assert received == expected

    def test_no_virtual_env(self):
        path = Path("/not/a/venv/bin/python3")
        expected = None, False
        received = utils.get_venv_like_prefix(str(path))
        assert received == expected


def compare_python_paths(path1, path2):
    return path1.parent == path2.parent


@pytest.mark.path
def test_find_python_in_path(tmp_path):
    assert utils.find_python_in_path(sys.executable) == pathlib.Path(sys.executable).absolute()

    posix_path_to_executable = pathlib.Path(sys.executable)
    assert compare_python_paths(
        utils.find_python_in_path(sys.prefix),
        posix_path_to_executable,
    )

    assert not utils.find_python_in_path(tmp_path)


@pytest.mark.parametrize(
    "given,expected",
    [
        ("scheme://netloc/path@rev#fragment", "rev"),
        ("scheme://netloc/path@rev", "rev"),
        ("scheme://netloc/path", ""),
        ("scheme://netloc/path#fragment", ""),
    ],
)
def test_get_rev_from_url(given, expected):
    assert utils.get_rev_from_url(given) == expected


@pytest.mark.parametrize(
    "given,expected",
    [
        (("ProjectName", False), "ProjectName"),
        (("ProjectName", True), "projectname"),
        (("1Project_Name", False), "1Project-Name"),
        (("1Project_Name", True), "1project-name"),
        (("Project-Name", False), "Project-Name"),
        (("Project-Name", True), "project-name"),
        (("Project123Name", False), "Project123Name"),
        (("Project123name", True), "project123name"),
        (("123$!ProjectName", False), "123-ProjectName"),
        (("123$!ProjectName", True), "123-projectname"),
        (("123$!Project_Name", False), "123-Project-Name"),
        (("123$!Project_Name", True), "123-project-name"),
        (("$!123Project_Name4", False), "-123Project-Name4"),
        (("$!123Project_Name4", True), "-123project-name4"),
    ],
)
def test_normalize_name(given, expected):
    assert utils.normalize_name(*given) == expected


class TestIsEditable:
    @mock.patch("pdm.utils.is_egg_link", return_value=True)
    @mock.patch("pdm.compat.Distribution")
    def test_is_egg_link(self, distribution, is_egg_link_patch):
        with is_egg_link_patch:
            assert utils.is_editable(distribution) is True

    @mock.patch("pdm.utils.is_egg_link", return_value=False)
    @mock.patch("pdm.compat.Distribution")
    def test_not_direct_url_distribution(self, distribution, is_egg_link_patch):
        distribution.read_text.return_value = None
        with is_egg_link_patch:
            assert utils.is_editable(distribution) is False

    @mock.patch("pdm.utils.is_egg_link", return_value=False)
    @mock.patch("pdm.compat.Distribution")
    def test_direct_url_distribution(self, distribution, is_egg_link_patch):
        distribution.read_text.return_value = """{"dir_info": {"editable": true}}"""
        with is_egg_link_patch:
            assert utils.is_editable(distribution) is True


def test_merge_dictionary():
    target = tomlkit.item(
        {
            "existing_dict": {"foo": "bar", "hello": "world"},
            "existing_list": ["hello"],
        }
    )
    input_dict = {
        "existing_dict": {"foo": "baz"},
        "existing_list": ["world"],
        "new_dict": {"name": "Sam"},
    }
    cli_utils.merge_dictionary(target, input_dict)
    assert target == {
        "existing_dict": {"foo": "baz", "hello": "world"},
        "existing_list": ["hello", "world"],
        "new_dict": {"name": "Sam"},
    }


def setup_dependencies(project):
    project.pyproject.metadata.update(
        {
            "dependencies": ["requests"],
            "optional-dependencies": {"web": ["flask"], "auth": ["passlib"]},
        }
    )
    project.pyproject.dependency_groups.update({"test": ["pytest"], "doc": ["mkdocs"]})
    project.pyproject.write()


@pytest.mark.parametrize(
    "args,golden",
    [
        ({"default": True, "dev": None, "groups": ()}, ["default", "test", "doc"]),
        (
            {"default": True, "dev": None, "groups": [":all"]},
            ["default", "web", "auth", "test", "doc"],
        ),
        (
            {"default": True, "dev": True, "groups": ["web"]},
            ["default", "web", "test", "doc"],
        ),
        (
            {"default": True, "dev": None, "groups": ["web"]},
            ["default", "web", "test", "doc"],
        ),
        ({"default": True, "dev": None, "groups": ["test"]}, ["default", "test"]),
        (
            {"default": True, "dev": None, "groups": ["test", "web"]},
            ["default", "test", "web"],
        ),
        ({"default": True, "dev": False, "groups": ["web"]}, ["default", "web"]),
        ({"default": False, "dev": None, "groups": ()}, ["test", "doc"]),
    ],
)
def test_dependency_group_selection(project, args, golden):
    setup_dependencies(project)
    selection = GroupSelection(project, **args)
    assert sorted(golden) == sorted(selection)


@pytest.mark.parametrize(
    "args,golden",
    [
        ({"groups": [":all"], "excluded_groups": ["web"]}, ["default", "auth", "doc", "test"]),
        ({"groups": [":all"], "excluded_groups": ["web", "auth"]}, ["default", "doc", "test"]),
        ({"groups": [":all"], "excluded_groups": ["default", "test"]}, ["auth", "doc", "web"]),
    ],
)
def test_exclude_optional_groups_from_all(project, args, golden):
    setup_dependencies(project)
    selection = GroupSelection(project, **args)
    assert golden == list(selection)


def test_prod_should_not_be_with_dev(project):
    setup_dependencies(project)
    selection = GroupSelection(project, default=True, dev=False, groups=["test"])
    with pytest.raises(PdmUsageError):
        list(selection)


def test_deprecation_warning():
    with pytest.warns(PDMWarning) as record:
        utils.deprecation_warning("Test warning", raise_since="99.99")
    assert len(record) == 1
    assert str(record[0].message) == "Test warning"

    with pytest.raises(PDMWarning):
        utils.deprecation_warning("Test warning", raise_since="0.0")


def test_comparable_version():
    assert utils.comparable_version("1.2.3") == utils.parse_version("1.2.3")
    assert utils.comparable_version("1.2.3a1+local1") == utils.parse_version("1.2.3a1")


@pytest.mark.parametrize("name", ["foO", "f", "3d", "f3", "333", "foo.bar", "foo-bar", "foo_bar", "foo-_.bar"])
def test_validate_project_name(name):
    assert utils.validate_project_name(name)


@pytest.mark.parametrize("name", ["", "-", ".foo", "foo_", "-3", "foo$bar", "a.3-"])
def test_invalidate_project_name(name):
    assert not utils.validate_project_name(name)


@pytest.mark.parametrize(
    "given,sanitized",
    [("foo", "foo"), ("Foo.Bar", "Foo.Bar"), ("-foo", "foo"), ("Foo%$Bar", "Foo-Bar"), ("Foo$", "Foo")],
)
def test_sanitize_project_name(given, sanitized):
    assert utils.sanitize_project_name(given) == sanitized

    with pytest.raises(PdmException):
        utils.sanitize_project_name("@#$")


@pytest.mark.parametrize(
    "url, redacted",
    [
        ("http://user:pass@localhost:8000/path?query=1#fragment", "http://*****@localhost:8000/path?query=1#fragment"),
        ("https://example.com", "https://example.com"),
        ("ftp://ftp.example.com/file.txt", "ftp://ftp.example.com/file.txt"),
    ],
)
def test_hide_url(url, redacted):
    hidden = utils.hide_url(url)
    assert hidden.secret == url
    assert hidden.redacted == str(hidden) == redacted


def test_fs_supports_link_method_when_method_doesnt_exist():
    assert utils.fs_supports_link_method("nonexistent-link-method") is False


def test_convert_to_datetime_when_uppercase_t_is_present():
    original_datetime = "2025-10-30T10:26:29Z"
    expected_datetime = datetime.fromisoformat("2025-10-30T10:26:29+00:00")

    assert utils.convert_to_datetime(original_datetime) == expected_datetime


def test_convert_to_datetime_when_uppercase_t_is_absent():
    original_datetime = "2025-10-30"
    expected_datetime = datetime(2025, 10, 30, tzinfo=timezone.utc)

    assert utils.convert_to_datetime(original_datetime) == expected_datetime


@pytest.mark.parametrize(
    ("value", "expected_datetime"),
    [
        ("2d", datetime(2025, 10, 28, 12, 0, tzinfo=timezone.utc)),
        ("12h", datetime(2025, 10, 30, 0, 0, tzinfo=timezone.utc)),
        ("3w", datetime(2025, 10, 9, 12, 0, tzinfo=timezone.utc)),
    ],
)
def test_convert_to_datetime_when_relative_duration_is_given(monkeypatch, value, expected_datetime):
    frozen_now = datetime(2025, 10, 30, 12, 0, tzinfo=timezone.utc)

    class FrozenDateTime(datetime):
        @classmethod
        def now(cls, tz=None):
            assert tz is timezone.utc
            return frozen_now

    monkeypatch.setattr(utils, "datetime", FrozenDateTime)

    assert utils.convert_to_datetime(value) == expected_datetime



import json
import os
import pathlib
from unittest import mock

import pytest
from rich.box import ASCII

from pdm.cli.commands.list import Command
from pdm.models.specifiers import PySpecSet
from pdm.pytest import Distribution
from tests import FIXTURES


def test_list_command(project, pdm, mocker):
    # Calls the correct handler within the Command
    m = mocker.patch.object(Command, "handle_list")
    pdm(["list"], obj=project)
    m.assert_called_once()


@pytest.mark.usefixtures("working_set")
def test_list_graph_command(project, pdm, mocker):
    # Calls the correct handler within the list command
    m = mocker.patch.object(Command, "handle_graph")
    pdm(["list", "--tree"], obj=project)
    m.assert_called_once()


@mock.patch("rich.console.ConsoleOptions.ascii_only", lambda: True)
@pytest.mark.usefixtures("working_set")
def test_list_dependency_graph(project, pdm):
    # Shows a line that contains a sub requirement (any order).
    pdm(["add", "requests"], obj=project, strict=True)
    result = pdm(["list", "--tree"], obj=project)
    expected = "-- urllib3 1.22 [ required: <1.24,>=1.21.1 ]" in result.outputs
    assert expected, result.outputs


@mock.patch("rich.console.ConsoleOptions.ascii_only", lambda: True)
@pytest.mark.usefixtures("working_set")
def test_list_dependency_graph_include_exclude(project, pdm):
    # Just include dev packages in the graph
    project.environment.python_requires = PySpecSet(">=3.6")
    dep_path = FIXTURES.joinpath("projects/demo").as_posix()
    pdm(["add", "-de", f"{dep_path}[security]"], obj=project, strict=True)

    # Output full graph
    result = pdm(["list", "--tree"], obj=project)
    expects = (
        "demo 0.0.1 [ Not required ]\n",
        "+-- chardet 3.0.4 [ required: Any ]\n" if os.name == "nt" else "",
        "`-- idna 2.7 [ required: Any ]\n",
        "requests 2.19.1 [ Not required ]\n",
        "+-- certifi 2018.11.17 [ required: >=2017.4.17 ]\n",
        "+-- chardet 3.0.4 [ required: <3.1.0,>=3.0.2 ]\n",
        "+-- idna 2.7 [ required: <2.8,>=2.5 ]\n",
        "`-- urllib3 1.22 [ required: <1.24,>=1.21.1 ]\n",
    )
    expects = "".join(expects)
    assert expects == result.outputs

    # Now exclude the dev dep.
    result = pdm(["list", "--tree", "--exclude", "dev"], obj=project)
    expects = ""
    assert expects == result.outputs

    # Only include the dev dep
    result = pdm(["list", "--tree", "--include", "dev", "--exclude", "*"], obj=project)
    expects = "demo[security] 0.0.1 [ required: Any ]\n"
    expects = "".join(expects)
    assert expects == result.outputs


@pytest.mark.usefixtures("working_set")
def test_list_dependency_graph_with_circular_forward(project, pdm, repository):
    # shows a circular dependency
    repository.add_candidate("foo", "0.1.0")
    repository.add_candidate("foo-bar", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["foo-bar"])
    repository.add_dependencies("foo-bar", "0.1.0", ["foo"])
    pdm(["add", "foo"], obj=project, strict=True)
    result = pdm(["list", "--tree"], obj=project)
    circular_found = "foo [circular]" in result.outputs
    assert circular_found


@mock.patch("rich.console.ConsoleOptions.ascii_only", lambda: True)
@pytest.mark.usefixtures("working_set")
def test_list_dependency_graph_with_circular_reverse(project, pdm, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_candidate("foo-bar", "0.1.0")
    repository.add_candidate("baz", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["foo-bar"])
    repository.add_dependencies("foo-bar", "0.1.0", ["foo", "baz"])
    repository.add_dependencies("baz", "0.1.0", [])
    pdm(["add", "foo"], obj=project, strict=True)

    # --reverse flag shows packages reversed and with [circular]
    result = pdm(["list", "--tree", "--reverse"], obj=project)
    expected = (
        "baz 0.1.0 \n"
        "`-- foo-bar 0.1.0 [ requires: Any ]\n"
        "    `-- foo 0.1.0 [ requires: Any ]\n"
        "        +-- foo-bar [circular] [ requires: Any ]\n"
        "        `-- test-project 0.0.0 [ requires: >=0.1.0 ]\n"
    )
    assert expected in result.outputs

    # -r flag shows packages reversed and with [circular]
    result = pdm(["list", "--tree", "-r"], obj=project)
    assert expected in result.outputs


def test_list_reverse_without_graph_flag(project, pdm):
    # results in PDMUsageError since --reverse needs --tree
    result = pdm(["list", "--reverse"], obj=project)
    assert "[PdmUsageError]" in result.stderr
    assert "--reverse cannot be used without --tree" in result.stderr

    result = pdm(["list", "-r"], obj=project)
    assert "[PdmUsageError]" in result.stderr
    assert "--reverse cannot be used without --tree" in result.stderr


@mock.patch("rich.console.ConsoleOptions.ascii_only", lambda: True)
@pytest.mark.usefixtures("working_set")
def test_list_reverse_dependency_graph(project, pdm):
    # requests visible on leaf node
    pdm(["add", "requests"], obj=project, strict=True)
    result = pdm(["list", "--tree", "--reverse"], obj=project)
    assert "`-- requests 2.19.1 [ requires: <1.24,>=1.21.1 ]" in result.outputs


@pytest.mark.usefixtures("working_set")
def test_list_json(project, pdm):
    # check json output matches graph output
    pdm(["add", "requests", "--no-self"], obj=project, strict=True)
    result = pdm(["list", "--tree", "--json"], obj=project)

    expected = [
        {
            "package": "requests",
            "version": "2.19.1",
            "required": ">=2.19.1",
            "dependencies": [
                {
                    "package": "certifi",
                    "version": "2018.11.17",
                    "required": ">=2017.4.17",
                    "dependencies": [],
                },
                {
                    "package": "chardet",
                    "version": "3.0.4",
                    "required": "<3.1.0,>=3.0.2",
                    "dependencies": [],
                },
                {
                    "package": "idna",
                    "version": "2.7",
                    "required": "<2.8,>=2.5",
                    "dependencies": [],
                },
                {
                    "package": "urllib3",
                    "version": "1.22",
                    "required": "<1.24,>=1.21.1",
                    "dependencies": [],
                },
            ],
        }
    ]
    assert expected == json.loads(result.outputs)


@pytest.mark.usefixtures("working_set")
def test_list_json_with_pattern(project, pdm):
    pdm(["add", "requests", "--no-self"], obj=project, strict=True)
    result = pdm(["list", "--tree", "--json", "chardet"], obj=project)

    expected = [
        {
            "package": "chardet",
            "version": "3.0.4",
            "required": "<3.1.0,>=3.0.2",
            "dependencies": [],
        },
    ]
    assert expected == json.loads(result.outputs)


@pytest.mark.usefixtures("working_set")
def test_list_json_reverse(project, pdm):
    # check json output matches reversed graph
    pdm(["add", "requests", "--no-self"], obj=project, strict=True)
    result = pdm(["list", "--tree", "--reverse", "--json"], obj=project)
    expected = [
        {
            "package": "certifi",
            "version": "2018.11.17",
            "requires": None,
            "dependents": [
                {
                    "package": "requests",
                    "version": "2.19.1",
                    "requires": ">=2017.4.17",
                    "dependents": [],
                }
            ],
        },
        {
            "package": "chardet",
            "version": "3.0.4",
            "requires": None,
            "dependents": [
                {
                    "package": "requests",
                    "version": "2.19.1",
                    "requires": "<3.1.0,>=3.0.2",
                    "dependents": [],
                }
            ],
        },
        {
            "package": "idna",
            "version": "2.7",
            "requires": None,
            "dependents": [
                {
                    "package": "requests",
                    "version": "2.19.1",
                    "requires": "<2.8,>=2.5",
                    "dependents": [],
                }
            ],
        },
        {
            "package": "urllib3",
            "version": "1.22",
            "requires": None,
            "dependents": [
                {
                    "package": "requests",
                    "version": "2.19.1",
                    "requires": "<1.24,>=1.21.1",
                    "dependents": [],
                }
            ],
        },
    ]

    assert expected == json.loads(result.outputs)


@pytest.mark.usefixtures("working_set")
def test_list_reverse_json_with_pattern(project, pdm):
    # check json output matches reversed graph
    pdm(["add", "requests", "--no-self"], obj=project, strict=True)
    result = pdm(["list", "--tree", "--reverse", "--json", "certifi"], obj=project)
    expected = [
        {
            "package": "certifi",
            "version": "2018.11.17",
            "requires": None,
            "dependents": [
                {
                    "package": "requests",
                    "version": "2.19.1",
                    "requires": ">=2017.4.17",
                    "dependents": [],
                }
            ],
        },
    ]

    assert expected == json.loads(result.outputs)


@pytest.mark.usefixtures("working_set")
def test_list_json_with_circular_forward(project, pdm, repository):
    # circulars are handled in json exports
    repository.add_candidate("foo", "0.1.0")
    repository.add_candidate("foo-bar", "0.1.0")
    repository.add_candidate("baz", "0.1.0")
    repository.add_dependencies("baz", "0.1.0", ["foo"])
    repository.add_dependencies("foo", "0.1.0", ["foo-bar"])
    repository.add_dependencies("foo-bar", "0.1.0", ["foo"])
    pdm(["add", "baz", "--no-self"], obj=project, strict=True)
    result = pdm(["list", "--tree", "--json"], obj=project)
    expected = [
        {
            "package": "baz",
            "version": "0.1.0",
            "required": ">=0.1.0",
            "dependencies": [
                {
                    "package": "foo",
                    "version": "0.1.0",
                    "required": "Any",
                    "dependencies": [
                        {
                            "package": "foo-bar",
                            "version": "0.1.0",
                            "required": "Any",
                            "dependencies": [
                                {
                                    "package": "foo",
                                    "version": "0.1.0",
                                    "required": "Any",
                                    "dependencies": [],
                                }
                            ],
                        }
                    ],
                }
            ],
        },
    ]
    assert expected == json.loads(result.outputs)


@pytest.mark.usefixtures("working_set")
def test_list_json_with_circular_reverse(project, pdm, repository):
    # circulars are handled in reversed json exports
    repository.add_candidate("foo", "0.1.0")
    repository.add_candidate("foo-bar", "0.1.0")
    repository.add_candidate("baz", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["foo-bar"])
    repository.add_dependencies("foo-bar", "0.1.0", ["foo", "baz"])
    repository.add_dependencies("baz", "0.1.0", [])
    pdm(["add", "foo", "--no-self"], obj=project, strict=True)
    result = pdm(["list", "--tree", "--reverse", "--json"], obj=project)
    expected = [
        {
            "package": "baz",
            "version": "0.1.0",
            "requires": None,
            "dependents": [
                {
                    "package": "foo-bar",
                    "version": "0.1.0",
                    "requires": "Any",
                    "dependents": [
                        {
                            "package": "foo",
                            "version": "0.1.0",
                            "requires": "Any",
                            "dependents": [
                                {
                                    "package": "foo-bar",
                                    "version": "0.1.0",
                                    "requires": "Any",
                                    "dependents": [],
                                }
                            ],
                        }
                    ],
                }
            ],
        },
    ]
    assert expected == json.loads(result.outputs)


def test_list_field_unknown(project, pdm):
    # unknown list fields flagged to user
    result = pdm(["list", "--fields", "notvalid"], obj=project)
    assert "[PdmUsageError]" in result.stderr
    assert "--fields must specify one or more of:" in result.stderr


def test_list_sort_unknown(project, pdm):
    # unknown sort fields flagged to user
    result = pdm(["list", "--sort", "notvalid"], obj=project)
    assert "[PdmUsageError]" in result.stderr
    assert "--sort key must be one of:" in result.stderr


def test_list_freeze_banned_options(project, pdm):
    # other flags cannot be used with --freeze
    result = pdm(["list", "--freeze", "--tree"], obj=project)
    expected = "--tree cannot be used with --freeze"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--reverse"], obj=project)
    expected = "--reverse cannot be used without --tree"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "-r"], obj=project)
    expected = "--reverse cannot be used without --tree"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--fields", "name"], obj=project)
    expected = "--fields cannot be used with --freeze"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--resolve"], obj=project)
    expected = "--resolve cannot be used with --freeze"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--sort", "version"], obj=project)
    expected = "--sort cannot be used with --freeze"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--csv"], obj=project)
    expected = "--csv cannot be used with --freeze"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--json"], obj=project)
    expected = "--json cannot be used with --freeze"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--markdown"], obj=project)
    expected = "--markdown cannot be used with --freeze"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--include", "dev"], obj=project)
    expected = "--include/--exclude cannot be used with --freeze"
    assert expected in result.outputs

    result = pdm(["list", "--freeze", "--exclude", "dev"], obj=project)
    expected = "--include/--exclude cannot be used with --freeze"
    assert expected in result.outputs


def test_list_multiple_export_formats(project, pdm):
    # export formats cannot be used with each other
    result = pdm(["list", "--csv", "--markdown"], obj=project)
    expected = "--markdown: not allowed with argument --csv"
    assert expected in result.outputs

    result = pdm(["list", "--csv", "--json"], obj=project)
    expected = "--json: not allowed with argument --csv"
    assert expected in result.outputs

    result = pdm(["list", "--markdown", "--csv"], obj=project)
    expected = "--csv: not allowed with argument --markdown"
    assert expected in result.outputs

    result = pdm(["list", "--markdown", "--json"], obj=project)
    expected = "--json: not allowed with argument --markdown"
    assert expected in result.outputs

    result = pdm(["list", "--json", "--markdown"], obj=project)
    expected = "--markdown: not allowed with argument --json"
    assert expected in result.outputs

    result = pdm(["list", "--json", "--csv"], obj=project)
    expected = "--csv: not allowed with argument --json"
    assert expected in result.outputs


@mock.patch("pdm.termui.ROUNDED", ASCII)
@pytest.mark.usefixtures("working_set")
def test_list_bare(project, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    result = pdm(["list"], obj=project)
    # Ordering can be different on different platforms
    # and python versions.
    assert "| name         | version    | location |\n" in result.output
    assert "| certifi      | 2018.11.17 |          |\n" in result.output
    assert "| chardet      | 3.0.4      |          |\n" in result.output
    assert "| idna         | 2.7        |          |\n" in result.output
    assert "| requests     | 2.19.1     |          |\n" in result.output
    assert "| urllib3      | 1.22       |          |\n" in result.output
    assert "| test-project | 0.0.0      |          |\n" in result.output


@mock.patch("pdm.termui.ROUNDED", ASCII)
@pytest.mark.usefixtures("working_set")
def test_list_bare_sorted_name(project, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    result = pdm(["list", "--sort", "name"], obj=project)
    expected = (
        "+--------------------------------------+\n"
        "| name         | version    | location |\n"
        "|--------------+------------+----------|\n"
        "| certifi      | 2018.11.17 |          |\n"
        "| chardet      | 3.0.4      |          |\n"
        "| idna         | 2.7        |          |\n"
        "| requests     | 2.19.1     |          |\n"
        "| test-project | 0.0.0      |          |\n"
        "| urllib3      | 1.22       |          |\n"
        "+--------------------------------------+\n"
    )
    assert expected == result.output


@mock.patch("pdm.termui.ROUNDED", ASCII)
@pytest.mark.usefixtures("working_set")
def test_list_with_pattern(project, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    result = pdm(["list", "--sort", "name", "c*"], obj=project)
    expected = (
        "+---------------------------------+\n"
        "| name    | version    | location |\n"
        "|---------+------------+----------|\n"
        "| certifi | 2018.11.17 |          |\n"
        "| chardet | 3.0.4      |          |\n"
        "+---------------------------------+\n"
    )
    assert expected == result.output


@pytest.fixture()
def fake_working_set(working_set):
    """Create fake packages with license data
    for testing.
    """

    class _MockPackagePath(pathlib.PurePosixPath):
        def read_text(self, *args, **kwargs):
            return self.license_text

    # Foo package (adapted to contain newlines in License field)
    # e.g. via license = { file="LICENSE" }
    foo = Distribution(
        "foo",
        "0.1.0",
        metadata={
            "License": "A License\n\nextra\ntext",
            "Classifier": "License :: A License",
        },
    )
    foo_l = _MockPackagePath("foo-0.1.0.dist-info", "LICENSE")
    foo_l.license_text = "license text for foo here"
    foo.files = [foo_l]

    # Bar package.
    bar = Distribution("bar", "3.0.1", metadata={"License": "B License"})
    bar_l = _MockPackagePath("bar-3.0.1.dist-info", "LICENSE")
    bar_l.license_text = "license text for bar here"
    bar.files = [bar_l]

    # Baz package.
    baz = Distribution("baz", "2.7", metadata={"License": "C License"})
    baz_l = _MockPackagePath("bar-2.7.dist-info", "LICENSE")
    baz_l.license_text = "license text for baz here"
    baz.files = [baz_l]

    # missing package- License is set to UNKNOWN, text saved in COPYING
    unknown = Distribution(
        "unknown",
        "1.0",
        metadata={
            "License": "UNKNOWN",
            "Classifier": "License :: OSI Approved :: Apache Software License",
        },
    )
    unknown_l = _MockPackagePath("unknown-1.0.dist-info", "COPYING")
    unknown_l.license_text = "license text for UNKNOWN here"
    unknown.files = [unknown_l]

    # missing package- License is set to UNKNOWN, text saved in LICENCE
    # using UK spelling
    classifier = Distribution("classifier", "1.0", metadata={"Classifier": "License :: PDM TEST D"})
    classifier_l = _MockPackagePath("classifier-1.0.dist-info", "LICENCE")
    classifier_l.license_text = "license text for CLASSIFIER here"
    classifier_l.read_text = lambda *a, **kw: 1 / 0  # make it throw an error
    classifier.files = [classifier_l]

    # Place our fake packages in the working set.
    for candidate in [foo, bar, baz, unknown, classifier]:
        working_set.add_distribution(candidate)
    return working_set


@pytest.fixture()
def fake_metadata(mocker, repository):
    def prepare_metadata(self):
        can = self.candidate
        version, dependencies = repository.get_raw_dependencies(can)
        dist = Distribution(can.name, version, can.req.editable)
        dist.dependencies = dependencies
        return dist

    return mocker.patch(
        "pdm.models.candidates.PreparedCandidate.prepare_metadata",
        prepare_metadata,
    )


@mock.patch("pdm.termui.ROUNDED", ASCII)
@pytest.mark.usefixtures("working_set")
def test_list_freeze(project, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    result = pdm(["list", "--freeze"], obj=project)
    expected = "certifi==2018.11.17\nchardet==3.0.4\nidna==2.7\nrequests==2.19.1\ntest-project==0.0.0\nurllib3==1.22\n"
    assert expected == result.output


@mock.patch("pdm.termui.ROUNDED", ASCII)
@pytest.mark.usefixtures("working_set")
def test_list_bare_sorted_version(project, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    result = pdm(["list", "--sort", "version"], obj=project)
    expected = (
        "+--------------------------------------+\n"
        "| name         | version    | location |\n"
        "|--------------+------------+----------|\n"
        "| test-project | 0.0.0      |          |\n"
        "| urllib3      | 1.22       |          |\n"
        "| requests     | 2.19.1     |          |\n"
        "| idna         | 2.7        |          |\n"
        "| certifi      | 2018.11.17 |          |\n"
        "| chardet      | 3.0.4      |          |\n"
        "+--------------------------------------+\n"
    )
    assert expected == result.output


@mock.patch("pdm.termui.ROUNDED", ASCII)
@pytest.mark.usefixtures("fake_metadata")
def test_list_bare_sorted_version_resolve(project, pdm, working_set):
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "requests", "--no-sync"], obj=project, strict=True)

    result = pdm(["list", "--sort", "version", "--resolve"], obj=project, strict=True)
    assert "requests" not in working_set
    expected = (
        "+----------------------------------+\n"
        "| name     | version    | location |\n"
        "|----------+------------+----------|\n"
        "| urllib3  | 1.22       |          |\n"
        "| requests | 2.19.1     |          |\n"
        "| idna     | 2.7        |          |\n"
        "| certifi  | 2018.11.17 |          |\n"
        "| chardet  | 3.0.4      |          |\n"
        "+----------------------------------+\n"
    )
    assert expected == result.outputs, result.outputs


@mock.patch("pdm.termui.ROUNDED", ASCII)
@pytest.mark.usefixtures("fake_working_set")
def test_list_bare_fields_licences(project, pdm):
    result = pdm(["list", "--fields", "name,version,groups,licenses"], obj=project)
    expected = (
        "+---------------------------------------------------------+\n"
        "| name       | version | groups | licenses                |\n"
        "|------------+---------+--------+-------------------------|\n"
        "| bar        | 3.0.1   | :sub   | B License               |\n"
        "| baz        | 2.7     | :sub   | C License               |\n"
        "| classifier | 1.0     | :sub   | PDM TEST D              |\n"
        "| foo        | 0.1.0   | :sub   | A License               |\n"
        "| unknown    | 1.0     | :sub   | Apache Software License |\n"
        "+---------------------------------------------------------+\n"
    )
    assert expected == result.output


@pytest.mark.usefixtures("fake_working_set")
def test_list_csv_fields_licences(project, pdm):
    result = pdm(["list", "--csv", "--fields", "name,version,licenses"], obj=project)
    expected = (
        "name,version,licenses\n"
        "bar,3.0.1,B License\n"
        "baz,2.7,C License\n"
        "classifier,1.0,PDM TEST D\n"
        "foo,0.1.0,A License\n"
        "unknown,1.0,Apache Software License\n"
    )
    assert expected == result.output


@pytest.mark.usefixtures("fake_working_set")
def test_list_json_fields_licences(project, pdm):
    result = pdm(["list", "--json", "--fields", "name,version,licenses"], obj=project)
    expected = [
        {"name": "bar", "version": "3.0.1", "licenses": "B License"},
        {"name": "baz", "version": "2.7", "licenses": "C License"},
        {"name": "classifier", "version": "1.0", "licenses": "PDM TEST D"},
        {"name": "foo", "version": "0.1.0", "licenses": "A License"},
        {"name": "unknown", "version": "1.0", "licenses": "Apache Software License"},
    ]

    assert expected == json.loads(result.outputs)


@pytest.mark.usefixtures("fake_working_set")
def test_list_markdown_fields_licences(project, pdm):
    # Note that in "foo" the "License" metadata field ("License": "A License\n\nextra\ntext")
    # is ignored, in favour of the classifier and the LICENSE file.
    # This behaviour could be improved.
    result = pdm(["list", "--markdown", "--fields", "name,version,licenses"], obj=project)
    expected = (
        "# test-project licenses\n"
        "## bar\n\n"
        "| Name | bar |\n"
        "|----|----|\n"
        "| Version | 3.0.1 |\n"
        "| Licenses | B License |\n\n"
        "bar-3.0.1.dist-info/LICENSE\n\n\n"
        "````\n"
        "license text for bar here\n"
        "````\n\n\n"
        "## baz\n\n"
        "| Name | baz |\n"
        "|----|----|\n"
        "| Version | 2.7 |\n"
        "| Licenses | C License |\n\n"
        "bar-2.7.dist-info/LICENSE\n\n\n"
        "````\n"
        "license text for baz here\n"
        "````\n\n\n"
        "## classifier\n\n"
        "| Name | classifier |\n"
        "|----|----|\n"
        "| Version | 1.0 |\n"
        "| Licenses | PDM TEST D |\n\n"
        "classifier-1.0.dist-info/LICENCE\n\n\n"
        "````\n"
        "Problem finding license text: division by zero\n"
        "````\n\n\n"
        "## foo\n\n"
        "| Name | foo |\n"
        "|----|----|\n"
        "| Version | 0.1.0 |\n"
        "| Licenses | A License |\n\n"
        "foo-0.1.0.dist-info/LICENSE\n\n\n"
        "````\n"
        "license text for foo here\n"
        "````\n\n\n"
        "## unknown\n\n"
        "| Name | unknown |\n"
        "|----|----|\n"
        "| Version | 1.0 |\n"
        "| Licenses | Apache Software License |\n\n"
        "unknown-1.0.dist-info/COPYING\n\n\n"
        "````\n"
        "license text for UNKNOWN here\n"
        "````\n\n\n"
    )
    assert expected == result.output


@pytest.mark.usefixtures("working_set", "repository")
def test_list_csv_include_exclude_valid(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    dep_path = FIXTURES.joinpath("projects/demo").as_posix()
    pdm(["add", "-de", f"{dep_path}[security]"], obj=project, strict=True)
    result = pdm(
        [
            "list",
            "--csv",
            "--fields",
            "name,version,groups",
            "--sort",
            "name",
            "--include",
            "notexisting",
        ],
        obj=project,
    )
    assert "[PdmUsageError]" in result.outputs
    assert "--include groups must be selected from" in result.outputs
    assert "dev" in result.outputs
    assert "default" in result.outputs
    assert ":sub" in result.outputs


@pytest.mark.usefixtures("local_finder")
def test_list_packages_in_given_venv(project, pdm):
    project.pyproject.metadata["requires-python"] = ">=3.7"
    project.pyproject.write()
    project.global_config["python.use_venv"] = True
    pdm(["venv", "create"], obj=project, strict=True)
    pdm(["venv", "create", "--name", "second"], obj=project, strict=True)
    project._saved_python = None
    pdm(["add", "first", "--no-self"], obj=project, strict=True)
    second_lockfile = str(project.root / "pdm.2.lock")
    pdm(
        ["add", "-G", "second", "--no-self", "-L", second_lockfile, "--venv", "second", "editables"],
        obj=project,
        strict=True,
    )
    project.environment = None
    result1 = pdm(["list", "--freeze"], obj=project, strict=True)
    result2 = pdm(["list", "--freeze", "--venv", "second"], obj=project, strict=True)
    assert result1.output.strip() == "first==2.0.2"
    assert result2.output.strip() == "editables==0.2"


@pytest.mark.usefixtures("working_set", "repository")
def test_list_csv_include_exclude(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    dep_path = FIXTURES.joinpath("projects/demo").as_posix()
    pdm(["add", "-de", f"{dep_path}[security]"], obj=project, strict=True)

    # Show all groups.
    result = pdm(
        ["list", "--csv", "--fields", "name,version,groups", "--sort", "name"],
        obj=project,
    )
    expected = (
        "name,version,groups\n"
        "certifi,2018.11.17,:sub\n"
        "chardet,3.0.4,:sub\n"
        "demo,0.0.1,dev\n"
        "idna,2.7,:sub\n"
        "requests,2.19.1,:sub\n"
        "urllib3,1.22,:sub\n"
    )
    assert expected == result.output

    # Sub always included.
    result = pdm(
        [
            "list",
            "--csv",
            "--fields",
            "name,groups",
            "--sort",
            "name",
            "--include",
            "dev",
        ],
        obj=project,
    )
    expected = "name,groups\ncertifi,:sub\nchardet,:sub\ndemo,dev\nidna,:sub\nrequests,:sub\nurllib3,:sub\n"
    assert expected == result.output

    # Include all (default) except sub
    result = pdm(
        [
            "list",
            "--csv",
            "--fields",
            "name,groups",
            "--sort",
            "name",
            "--exclude",
            ":sub",
        ],
        obj=project,
    )
    expected = "name,groups\ndemo,dev\n"
    assert expected == result.output

    # Show just the dev group
    result = pdm(
        [
            "list",
            "--csv",
            "--fields",
            "name,version,groups",
            "--sort",
            "name",
            "--include",
            "dev",
            "--exclude",
            "*",
        ],
        obj=project,
    )
    expected = "name,version,groups\ndemo,0.0.1,dev\n"
    assert expected == result.output

    # Exclude the dev group.
    result = pdm(
        [
            "list",
            "--csv",
            "--fields",
            "name,version,groups",
            "--sort",
            "name",
            "--exclude",
            "dev",
        ],
        obj=project,
    )
    expected = "name,version,groups\n"
    assert expected == result.output



import os

import pytest

from pdm.cli.templates import ProjectTemplate
from pdm.exceptions import PdmException


def test_non_pyproject_template_disallowed(project_no_init):
    with ProjectTemplate("tests/fixtures/projects/demo_extras") as template:
        with pytest.raises(PdmException, match=r"Template pyproject.toml not found"):
            template.generate(project_no_init.root, {"project": {"name": "foo"}})


def test_module_project_template(project_no_init):
    metadata = {
        "project": {"name": "foo", "version": "0.1.0", "requires-python": ">=3.10"},
        "build-system": {"requires": ["pdm-backend"], "build-backend": "pdm.backend"},
    }

    with ProjectTemplate("tests/fixtures/projects/demo") as template:
        template.generate(project_no_init.root, metadata)

    project_no_init.pyproject.reload()
    assert project_no_init.pyproject.metadata["name"] == "foo"
    assert project_no_init.pyproject.metadata["requires-python"] == ">=3.10"
    assert project_no_init.pyproject._data["build-system"] == metadata["build-system"]
    assert project_no_init.pyproject.metadata["dependencies"] == ["idna", "chardet; os_name=='nt'"]
    assert project_no_init.pyproject.metadata["optional-dependencies"]["tests"] == ["pytest"]
    assert (project_no_init.root / "foo.py").exists()
    assert os.access(project_no_init.root / "foo.py", os.W_OK)


def test_module_project_template_generate_application(project_no_init):
    metadata = {
        "project": {"name": "", "version": "", "requires-python": ">=3.10"},
    }

    with ProjectTemplate("tests/fixtures/projects/demo") as template:
        template.generate(project_no_init.root, metadata)

    project_no_init.pyproject.reload()
    assert project_no_init.pyproject.metadata["name"] == ""
    assert "build-system" not in project_no_init.pyproject._data
    assert project_no_init.pyproject.metadata["dependencies"] == ["idna", "chardet; os_name=='nt'"]
    assert (project_no_init.root / "demo.py").exists()


def test_package_project_template(project_no_init):
    metadata = {
        "project": {"name": "foo", "version": "0.1.0", "requires-python": ">=3.10"},
        "build-system": {"requires": ["pdm-backend"], "build-backend": "pdm.backend"},
    }

    with ProjectTemplate("tests/fixtures/projects/demo-package") as template:
        template.generate(project_no_init.root, metadata)

    project_no_init.pyproject.reload()
    assert project_no_init.pyproject.metadata["name"] == "foo"
    assert project_no_init.pyproject.metadata["requires-python"] == ">=3.10"
    assert project_no_init.pyproject._data["build-system"] == metadata["build-system"]
    assert (project_no_init.root / "foo").is_dir()
    assert (project_no_init.root / "foo/__init__.py").exists()
    assert project_no_init.pyproject.settings["version"] == {"path": "foo/__init__.py", "source": "file"}



"""Additional tests for the info command to improve coverage"""

import json


def test_info_command_packages_option(project, pdm):
    """Test info command with --packages option"""
    result = pdm(["info", "--packages"], obj=project)
    assert result.exit_code == 0
    # Should show packages path
    assert result.output.strip() != ""


def test_info_command_all_options_mutually_exclusive(project, pdm):
    """Test that info command options are mutually exclusive"""
    # Only one field option should work at a time
    result = pdm(["info", "--python"], obj=project)
    assert result.exit_code == 0

    result = pdm(["info", "--where"], obj=project)
    assert result.exit_code == 0

    result = pdm(["info", "--packages"], obj=project)
    assert result.exit_code == 0

    result = pdm(["info", "--env"], obj=project)
    assert result.exit_code == 0


def test_info_command_env_output_format(project, pdm):
    """Test that --env outputs valid JSON"""
    result = pdm(["info", "--env"], obj=project, strict=True)
    # Try to parse as JSON
    try:
        data = json.loads(result.output)
        assert isinstance(data, dict)
    except json.JSONDecodeError:
        # If it's not JSON, it should at least contain marker info
        assert "python_version" in result.output or "implementation" in result.output


def test_info_command_json_contains_all_fields(project, pdm):
    """Test that --json output contains all expected fields"""
    result = pdm(["info", "--json"], obj=project, strict=True)

    data = json.loads(result.output)

    # Check pdm section
    assert "pdm" in data
    assert "version" in data["pdm"]

    # Check python section
    assert "python" in data
    assert "interpreter" in data["python"]
    assert "version" in data["python"]
    assert "markers" in data["python"]
    assert isinstance(data["python"]["markers"], dict)

    # Check project section
    assert "project" in data
    assert "root" in data["project"]
    assert "pypackages" in data["project"]


def test_info_command_default_output(project, pdm):
    """Test info command with no options shows formatted output"""
    result = pdm(["info"], obj=project, strict=True)

    # Should contain all major sections
    assert "PDM version" in result.output or "PDM" in result.output
    assert "Python Interpreter" in result.output or "Interpreter" in result.output
    assert "Project Root" in result.output or "Root" in result.output
    assert "Local Packages" in result.output or "Packages" in result.output


def test_info_command_with_global_project(pdm, tmp_path, monkeypatch):
    """Test info command with global project"""
    monkeypatch.chdir(tmp_path)

    result = pdm(["info", "-g", "--python"])
    assert result.exit_code == 0


def test_info_command_shows_venv_info(project, pdm):
    """Test info command shows virtual environment information when applicable"""
    # Create a venv
    project.global_config["python.use_venv"] = True

    result = pdm(["info"], obj=project)
    assert result.exit_code == 0


def test_info_command_non_local_environment(project, pdm, mocker):
    """Test info command with non-local environment shows site-packages"""
    # Mock the environment to be non-local
    project.environment.is_local = False

    # Mock get_paths to return purelib
    project.environment.get_paths = mocker.Mock(return_value={"purelib": "/fake/purelib"})

    result = pdm(["info", "--packages"], obj=project)
    assert result.exit_code == 0
    assert "/fake/purelib" in result.output or "purelib" in result.output


def test_info_command_global_project_prefix(project, pdm):
    """Test that global project shows 'Global' prefix in output"""
    result = pdm(["info", "-g"], obj=project)

    # If it's a global project, should show "Global" prefix
    # This test checks if the code path exists
    assert result.exit_code == 0



from __future__ import annotations

from collections.abc import Iterator

from resolvelib.resolvers import RequirementInformation

from pdm.models.candidates import Candidate
from pdm.models.requirements import parse_requirement
from pdm.resolver.providers import _CONFLICT_PRIORITY_THRESHOLD


def _build_candidates(identifier: str) -> dict[str, Iterator[Candidate]]:
    requirement = parse_requirement(identifier)
    candidate = Candidate(requirement, name=requirement.project_name, version="1.0")
    return {identifier: iter([candidate])}


def _build_information(identifier: str) -> dict[str, Iterator[RequirementInformation]]:
    requirement = parse_requirement(identifier)
    return {identifier: iter([RequirementInformation(requirement, None)])}


def test_narrow_requirement_selection_promotes_repeated_conflicts(project, repository):
    repository.add_candidate("conflict-pkg", "1.0")
    repository.add_candidate("other-pkg", "1.0")

    provider = project.get_provider()
    narrow = provider.narrow_requirement_selection
    causes = [RequirementInformation(parse_requirement("conflict-pkg"), None)]

    for _ in range(1, _CONFLICT_PRIORITY_THRESHOLD):
        result = list(narrow(["other-pkg"], {}, {}, {}, causes))
        assert result == ["other-pkg"]

    result = list(narrow(["other-pkg", "conflict-pkg"], {}, {}, {}, causes))
    assert result == ["conflict-pkg"]

    result = list(narrow(["other-pkg", "conflict-pkg"], {}, {}, {}, []))
    assert result == ["conflict-pkg"]

    other_causes = [RequirementInformation(parse_requirement("other-pkg"), None)]
    result = list(narrow(["other-pkg", "conflict-pkg"], {}, {}, {}, other_causes))
    assert result == ["other-pkg"]


def test_get_preference_prioritizes_promoted_conflicts(project, repository):
    repository.add_candidate("promoted-pkg", "1.0")
    repository.add_candidate("normal-pkg", "1.0")

    provider = project.get_provider()
    provider._conflict_promoted.add("promoted-pkg")

    promoted_preference = provider.get_preference(
        "promoted-pkg",
        {},
        _build_candidates("promoted-pkg"),
        _build_information("promoted-pkg"),
        [],
    )
    normal_preference = provider.get_preference(
        "normal-pkg",
        {},
        _build_candidates("normal-pkg"),
        _build_information("normal-pkg"),
        [],
    )

    assert promoted_preference < normal_preference



import shlex
import sys
from collections import namedtuple
from textwrap import dedent

import pytest

from pdm.cli import actions
from pdm.cli.options import from_splitted_env
from pdm.signals import pdm_signals

pytestmark = pytest.mark.usefixtures("repository", "working_set", "local_finder")


def test_pre_script_fail_fast(project, pdm, capfd, mocker):
    project.pyproject.settings["scripts"] = {
        "pre_install": "python -c \"print('PRE INSTALL CALLED'); exit(1)\"",
        "post_install": "python -c \"print('POST INSTALL CALLED')\"",
    }
    project.pyproject.write()
    synchronize = mocker.patch("pdm.installers.synchronizers.Synchronizer.synchronize")
    result = pdm(["install"], obj=project)
    assert result.exit_code == 1
    out, _ = capfd.readouterr()
    assert "PRE INSTALL CALLED" in out
    assert "POST INSTALL CALLED" not in out
    synchronize.assert_not_called()


def test_pre_and_post_scripts(project, pdm, capfd, _echo):
    project.pyproject.settings["scripts"] = {
        "pre_script": "python echo.py pre_script",
        "post_script": "python echo.py post_script",
        "pre_test": "python echo.py pre_test",
        "test": "python echo.py test",
        "post_test": "python echo.py post_test",
        "pre_run": "python echo.py pre_run",
        "post_run": "python echo.py post_run",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    expected = dedent(
        """
        pre_run CALLED
        pre_script CALLED
        pre_test CALLED
        test CALLED
        post_test CALLED
        post_script CALLED
        post_run CALLED
        """
    ).strip()
    assert out.strip() == expected


def test_composite_runs_all_hooks(project, pdm, capfd, _echo):
    project.pyproject.settings["scripts"] = {
        "test": {"composite": ["first", "second"]},
        "pre_test": "python echo.py Pre-Test",
        "post_test": "python echo.py Post-Test",
        "first": "python echo.py First",
        "pre_first": "python echo.py Pre-First",
        "second": "python echo.py Second",
        "post_second": "python echo.py Post-Second",
        "pre_script": "python echo.py Pre-Script",
        "post_script": "python echo.py Post-Script",
        "pre_run": "python echo.py Pre-Run",
        "post_run": "python echo.py Post-Run",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    expected = dedent(
        """
        Pre-Run CALLED
        Pre-Script CALLED
        Pre-Test CALLED
        Pre-Script CALLED
        Pre-First CALLED
        First CALLED
        Post-Script CALLED
        Pre-Script CALLED
        Second CALLED
        Post-Second CALLED
        Post-Script CALLED
        Post-Test CALLED
        Post-Script CALLED
        Post-Run CALLED
        """
    ).strip()
    assert out.strip() == expected


@pytest.mark.parametrize("option", [":all", ":pre,:post"])
def test_skip_all_hooks_option(project, pdm, capfd, option: str, _echo):
    project.pyproject.settings["scripts"] = {
        "test": {"composite": ["first", "second"]},
        "pre_test": "python echo.py Pre-Test",
        "post_test": "python echo.py Post-Test",
        "first": "python echo.py First",
        "pre_first": "python echo.py Pre-First",
        "post_first": "python echo.py Post-First",
        "second": "python echo.py Second",
        "pre_second": "python echo.py Pre-Second",
        "post_second": "python echo.py Post-Second",
        "pre_script": "python echo.py Pre-Script",
        "post_script": "python echo.py Post-Script",
        "pre_run": "python echo.py Pre-Run",
        "post_run": "python echo.py Post-Run",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", f"--skip={option}", "first"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-First CALLED" not in out
    assert "First CALLED" in out
    assert "Post-First CALLED" not in out
    assert "Pre-Script CALLED" not in out
    assert "Post-Script CALLED" not in out
    capfd.readouterr()
    pdm(["run", f"--skip={option}", "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-Test CALLED" not in out
    assert "Pre-First CALLED" not in out
    assert "First CALLED" in out
    assert "Post-First CALLED" not in out
    assert "Pre-Second CALLED" not in out
    assert "Second CALLED" in out
    assert "Post-Second CALLED" not in out
    assert "Post-Test CALLED" not in out
    assert "Pre-Script CALLED" not in out
    assert "Post-Script CALLED" not in out
    assert "Pre-Run CALLED" not in out
    assert "Post-Run CALLED" not in out


@pytest.mark.parametrize(
    "args",
    [
        "--skip pre_test,post_first,second",
        "-k pre_test,post_first,second",
        "--skip pre_test --skip post_first --skip second",
        "-k pre_test -k post_first -k second",
        "--skip pre_test --skip post_first,second",
        "-k pre_test -k post_first,second",
    ],
)
def test_skip_option(project, pdm, capfd, args, _echo):
    project.pyproject.settings["scripts"] = {
        "test": {"composite": ["first", "second"]},
        "pre_test": "python echo.py Pre-Test",
        "post_test": "python echo.py Post-Test",
        "first": "python echo.py First",
        "pre_first": "python echo.py Pre-First",
        "post_first": "python echo.py Post-First",
        "second": "python echo.py Second",
        "pre_second": "python echo.py Pre-Second",
        "post_second": "python echo.py Post-Second",
    }
    project.pyproject.write()
    capfd.readouterr()
    pdm(["run", *shlex.split(args), "test"], strict=True, obj=project)
    out, _ = capfd.readouterr()
    assert "Pre-Test CALLED" not in out
    assert "Pre-First CALLED" in out
    assert "First CALLED" in out
    assert "Post-First CALLED" not in out
    assert "Pre-Second CALLED" not in out
    assert "Second CALLED" not in out
    assert "Post-Second CALLED" not in out
    assert "Post-Test CALLED" in out


@pytest.mark.parametrize(
    "env, expected",
    [
        ("pre_test", ["pre_test"]),
        ("pre_test,post_test", ["pre_test", "post_test"]),
        ("pre_test , post_test", ["pre_test", "post_test"]),
        (None, None),
        (" ", None),
        (" , ", None),
    ],
)
def test_skip_option_default_from_env(env, expected, monkeypatch):
    if env is not None:
        monkeypatch.setenv("PDM_SKIP_HOOKS", env)

    # Default value is set once and not easily testable
    # so we test the function generating this default value
    assert from_splitted_env("PDM_SKIP_HOOKS", ",") == expected


HookSpecs = namedtuple("HookSpecs", ["command", "hooks", "fixtures"])
this_python_version = f"{sys.version_info[0]}.{sys.version_info[1]}"

KNOWN_COMMAND_HOOKS = (
    ("add", "add requests", ("pre_lock", "post_lock"), ["working_set"]),
    ("build", "build", ("pre_build", "post_build"), []),
    ("init", "init --non-interactive", ("post_init",), []),
    (
        "install",
        "install",
        ("pre_install", "post_install", "pre_lock", "post_lock"),
        ["repository"],
    ),
    ("lock", "lock", ("pre_lock", "post_lock"), []),
    (
        "publish",
        "publish --username abc --password 123",
        ("pre_publish", "pre_build", "post_build", "post_publish"),
        ["mock_publish"],
    ),
    ("remove", "remove requests", ("pre_lock", "post_lock"), ["lock"]),
    ("sync", "sync", ("pre_install", "post_install"), ["lock"]),
    ("update", "update", ("pre_install", "post_install", "pre_lock", "post_lock"), []),
    ("use", f"use -f {this_python_version}", ("post_use",), []),
)

parametrize_with_commands = pytest.mark.parametrize(
    "specs",
    [pytest.param(HookSpecs(command, hooks, fixtures), id=id) for id, command, hooks, fixtures in KNOWN_COMMAND_HOOKS],
)

parametrize_with_hooks = pytest.mark.parametrize(
    "specs,hook",
    [
        pytest.param(HookSpecs(command, hooks, fixtures), hook, id=f"{id}-{hook}")
        for id, command, hooks, fixtures in KNOWN_COMMAND_HOOKS
        for hook in hooks
    ],
)


@pytest.fixture
def hooked_project(project, capfd, specs, request):
    project.pyproject.settings["scripts"] = {hook: f"python -c \"print('{hook} CALLED')\"" for hook in pdm_signals}
    project.pyproject.write()
    for fixture in specs.fixtures:
        request.getfixturevalue(fixture)
    capfd.readouterr()
    return project


@pytest.fixture
def lock(project, capfd):
    project.add_dependencies(["requests"])
    actions.do_lock(project)
    capfd.readouterr()


@parametrize_with_commands
def test_hooks(hooked_project, pdm, capfd, specs: HookSpecs):
    pdm(shlex.split(specs.command), strict=True, obj=hooked_project)
    out, _ = capfd.readouterr()
    for hook in specs.hooks:
        assert f"{hook} CALLED" in out


@parametrize_with_hooks  # Iterate over hooks as we need a clean slate for each run
def test_skip_option_from_signal(hooked_project, pdm, capfd, specs: HookSpecs, hook: str):
    pdm([*shlex.split(specs.command), f"--skip={hook}"], strict=True, obj=hooked_project)
    out, _ = capfd.readouterr()
    assert f"{hook} CALLED" not in out
    for known_hook in specs.hooks:
        if known_hook != hook:
            assert f"{known_hook} CALLED" in out


@parametrize_with_commands
@pytest.mark.parametrize("option", [":all", ":pre,:post"])
def test_skip_all_option_from_signal(hooked_project, pdm, capfd, specs: HookSpecs, option: str):
    pdm(
        [*shlex.split(specs.command), f"--skip={option}"],
        strict=True,
        obj=hooked_project,
    )
    out, _ = capfd.readouterr()
    for hook in pdm_signals:
        assert f"{hook} CALLED" not in out


@parametrize_with_commands
@pytest.mark.parametrize("prefix", ["pre", "post"])
def test_skip_pre_post_option_from_signal(hooked_project, pdm, capfd, specs: HookSpecs, prefix: str):
    pdm(
        [*shlex.split(specs.command), f"--skip=:{prefix}"],
        strict=True,
        obj=hooked_project,
    )
    out, _ = capfd.readouterr()
    for hook in specs.hooks:
        if hook.startswith(prefix):
            assert f"{hook} CALLED" not in out
        else:
            assert f"{hook} CALLED" in out



"""Additional tests for the show command"""

import pytest

from pdm.cli.commands.show import filter_stable


def test_filter_stable_with_stable_version(mocker):
    """Test filter_stable returns True for stable versions"""
    # Mock package with stable version
    package = mocker.Mock()
    package.version = "1.0.0"
    assert filter_stable(package) is True


def test_filter_stable_with_prerelease_alpha(mocker):
    """Test filter_stable returns False for alpha prereleases"""
    package = mocker.Mock()
    package.version = "1.0.0a1"
    assert filter_stable(package) is False


def test_filter_stable_with_prerelease_beta(mocker):
    """Test filter_stable returns False for beta prereleases"""
    package = mocker.Mock()
    package.version = "2.0.0b2"
    assert filter_stable(package) is False


def test_filter_stable_with_prerelease_rc(mocker):
    """Test filter_stable returns False for release candidates"""
    package = mocker.Mock()
    package.version = "3.0.0rc1"
    assert filter_stable(package) is False


def test_filter_stable_with_dev_version(mocker):
    """Test filter_stable returns False for dev versions"""
    package = mocker.Mock()
    package.version = "1.0.0.dev1"
    assert filter_stable(package) is False


@pytest.mark.network
def test_show_command_with_specific_metadata_keys(pdm):
    """Test show command with specific metadata keys"""
    result = pdm(["show", "requests", "--name"])
    assert result.exit_code == 0
    assert "requests" in result.output.lower()

    result = pdm(["show", "requests", "--version"])
    assert result.exit_code == 0
    # Should contain a version number


@pytest.mark.network
def test_show_command_with_multiple_metadata_keys(pdm):
    """Test show command with multiple metadata keys only shows selected ones"""
    result = pdm(["show", "requests", "--name", "--version"])
    assert result.exit_code == 0
    # Should only show name and version, not full metadata


def test_show_command_non_distribution_project(project, pdm):
    """Test show command on a non-distribution project raises error"""
    # Mark the project as non-distribution
    project.pyproject.settings["distribution"] = False

    result = pdm(["show"], obj=project)
    assert result.exit_code != 0
    assert "not a library" in result.stderr
